from xicam.plugins import ProcessingPlugin, Input, Output
import numpy as np
from pyFAI import units
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator


# TODO Check if this moduel should be deleted as it duplicates qintegrate.py
class QIntegratePlugin(ProcessingPlugin):
    integrator = Input(description='A PyFAI.AzimuthalIntegrator object',
                       type=AzimuthalIntegrator)
    data = Input(description='2d array representing intensity for each pixel',
                 type=np.ndarray)
    npt = Input(description='Number of bins along q')
    polz_factor = Input(description='Polarization factor for correction',
                        type=float)
    unit = Input(description='Output units for q',
                 type=[str, units.Unit],
                 default="q_A^-1")
    radial_range = Input(description='The lower and upper range of the radial unit. If not provided, range is simply '
                                     '(data.min(), data.max()). Values outside the range are ignored.',
                         type=tuple)
    azimuth_range = Input(description='The lower and upper range of the azimuthal angle in degree. If not provided, '
                                      'range is simply (data.min(), data.max()). Values outside the range are ignored.')
    mask = Input(description='Array (same size as image) with 1 for masked pixels, and 0 for valid pixels',
                 type=np.ndarray)
    dark = Input(description='Dark noise image',
                 type=np.ndarray)
    flat = Input(description='Flat field image',
                 type=np.ndarray)
    method = Input(description='Can be "numpy", "cython", "BBox" or "splitpixel", "lut", "csr", "nosplit_csr", '
                               '"full_csr", "lut_ocl" and "csr_ocl" if you want to go on GPU. To Specify the device: '
                               '"csr_ocl_1,2"',
                   type=str)
    normalization_factor = Input(description='Value of a normalization monitor',
                                 type=float)
    q = Output(description='Q bin center positions',
               type=np.array)
    I = Output(description='Binned/pixel-split integrated intensity',
               type=np.array)

    def evaluate(self):
        self.q.value, self.I.value = self.integrator.value.integrate1d(data=self.data.value,
                                                                       npt=self.npt.value,
                                                                       radial_range=self.radial_range.value,
                                                                       azimuth_range=self.azimuth_range.value,
                                                                       mask=self.mask.value,
                                                                       polarization_factor=self.polz_factor.value,
                                                                       dark=self.dark.value,
                                                                       flat=self.flat.value,
                                                                       method=self.method.value,
                                                                       unit=self.unit.value,
                                                                       normalization_factor=self.normalization_factor.value)



from pyFAI import detectors
import numpy as np
from collections import OrderedDict


class FastCCD(detectors.Detector):
    aliases = ['Fast CCD']
    MAX_SHAPE = (960, 2050)

    def __init__(self, pixel1=30.e-6, pixel2=30.e-6, max_shape=MAX_SHAPE):
        super(FastCCD, self).__init__(pixel1=pixel1, pixel2=pixel2, max_shape=max_shape)

    def calc_mask(self):
        """
        Returns a generic mask for FastCCD detectors...
        """

        mask = np.zeros(MAX_SHAPE)
        # mask middle 19 cols
        # TODO: fix overscanned pixels at IOC instead of this 19 col mask here
        mask[:, 2050 // 2 - 9: 2050 // 2 + 9] = 1
        # TODO: check that this mask is correct

        return mask


class LAMBDA(detectors.Detector):
    aliases = ['LAMBDA']
    MAX_SHAPE = (1536, 512)

    def __init__(self, pixel1=22.e-6, pixel2=22.e-6, max_shape=MAX_SHAPE):
        super(LAMBDA, self).__init__(pixel1=pixel1, pixel2=pixel2, max_shape=max_shape)


if __name__ == '__main__':
    assert detectors.ALL_DETECTORS['fastccd']



import pytest


def test_VerticalCut():
    import numpy as np

    from qtpy.QtWidgets import QApplication
    app = QApplication([])
    from xicam.SAXS.operations.verticalcuts import VerticalCutPlugin
    t1 = VerticalCutPlugin()
    t1.data.value = np.ones((10, 10))
    t1.qz.value = np.tile(np.arange(10), (1, 10))
    t1.qzminimum.value = 3
    t1.qzmaximum.value = 6
    t1.evaluate()
    assert np.sum(t1.verticalcut.value) == 60



from typing import Callable

from pyqtgraph.parametertree import Parameter, ParameterTree
from pyqtgraph.parametertree.parameterTypes import ActionParameter, ListParameter

from ..workflows.xpcs import OneTimeAlgorithms, TwoTimeAlgorithms


class CorrelationParameterTree(ParameterTree):
    """Parameter tree that captures XPCS correlation adjustable parameters.

    Starting point to attempt to allow selecting an 'algorithm'
    and then repopulating the parameter tree accordingly.
    Any visible input parameters in the associated workflow will be adjustable here.
    The workflow can also be run here.
    """
    def __init__(self, parent=None, showHeader=True, processor: Callable[[], None] = None):
        """Create the parameter tree, optionally providing a workflow.

        When a workflow is provided,

        Parameters
        ----------
        parent
            See pyqtgraph.ParameterTree.__init__.
        showHeader
            See pyqtgraph.ParameterTree.__init__.
        processor
            If provided, creates a 'Run' button that is connected to the callable passed (default is None).
            For example, this callable could grab the necessary data to pass into a workflow, then execute it.
        """
        super(CorrelationParameterTree, self).__init__(parent, showHeader)
        self._paramName = 'Algorithm'
        self._name = 'Correlation Processor'
        self.processor = processor
        self.param = None
        self.workflow = None
        self._workflows = dict()

        # List of algorithms available
        # key   -> algorithm name (workflow.name)
        # value -> algorithm callable (workflow)
        self.listParameter = ListParameter(name=self._paramName,
                                           values={'':''},
                                           value='')

        self.param = Parameter(name=self._name)
        self.param.addChild(self.listParameter)
        self.setParameters(self.param, showTop=False)

        if self.processor:
            # Button added separately since update removes then adds all children in self.param
            self.processButton = ActionParameter(name="Run")
            self.processButton.sigActivated.connect(self.processor)
            self.addParameters(self.processButton)

    def update(self, *_):
        """Update the parameter tree according to which algorithm (workflow) is selected."""
        # for child in self.param.children():  # this doesn't seem to work...
        for child in self.param.childs[1:]:
            child.remove()

        # Based on current workflow (listParameter value), re-populate the tree.
        self.workflow = self._workflows.get(self.listParameter.value().name, self.listParameter.value()())
        self._workflows[self.workflow.name] = self.workflow
        for operation in self.workflow.operations:
            self.param.addChild(Parameter(name=operation.name, type='group', children=operation.as_parameter()))


class OneTimeParameterTree(CorrelationParameterTree):
    """Defines a parameter tree for 1-time correlation."""
    def __init__(self, parent=None, showHeader=True, processor: Callable[[], None] = None):
        super(OneTimeParameterTree, self).__init__(parent, showHeader, processor)
        self._name = '1-Time Processor'
        self.listParameter.setLimits(OneTimeAlgorithms.algorithms())
        self.listParameter.setValue(OneTimeAlgorithms.algorithms()[OneTimeAlgorithms.default()])

        self.update()
        self.listParameter.sigValueChanged.connect(self.update)


class TwoTimeParameterTree(CorrelationParameterTree):
    """Defines a parameter tree for 2-time correlation."""
    def __init__(self, parent=None, showHeader=True, processor: Callable[[], None] = None):
        super(TwoTimeParameterTree, self).__init__(parent, showHeader, processor)
        self._name = '2-Time Processor'
        self.listParameter.setLimits(TwoTimeAlgorithms.algorithms())
        self.listParameter.setValue(TwoTimeAlgorithms.algorithms()[TwoTimeAlgorithms.default()])

        self.update()
        self.listParameter.sigValueChanged.connect(self.update)


from camsaxs import cwt2d
from typing import Tuple
import numpy as np
from xicam.plugins.operationplugin import operation, output_names, display_name, describe_input, describe_output, \
    categories


@operation
@output_names('center')
@display_name("Ricker Wavelet Calibration")
@describe_input("data", 'Calibrant frame image data')
@describe_input("domain", 'Search domain in pixels: (r_min, r_max)')
@describe_input("width", 'Approximate width of the calibrant ring in pixels')
@describe_output("center", "The position in pixels of the approximated beam center")
@categories(("Scattering", "Calibration"))
def ricker_wavelet(data: np.ndarray, domain: Tuple[float], width: float = 5) -> Tuple[float]:
    center, _ = cwt2d(data, domain=domain, width=width)
    return center



import numpy as np
from dask import array as da
from typing import Tuple, List, Iterable
import pyqtgraph as pg

import skbeam.core.correlation as corr
from xicam.SAXS.patches.pyFAI import AzimuthalIntegrator
from xicam.core.intents import PlotIntent
from xicam.core import msg

from xicam.plugins.operationplugin import operation, describe_input, describe_output, visible, \
    input_names, output_names, display_name, categories, intent

from ..utils import get_label_array, average_q_from_labels


@operation
@display_name('1-time Correlation')
@input_names('images', 'labels', 'rois', 'image_item', 'number_of_buffers', 'number_of_levels',
             'intensity_drift_correction')
@describe_input('images', 'Input array of two or more dimensions')
@describe_input('labels', 'Labeled array of the same shape as the image stack. \
                 Each ROI is represented by sequential integers starting at one.  For \
                 example, if you have four ROIs, they must be labeled 1, 2, 3, 4. \
                 Background is labeled as 0')
@describe_input('number_of_buffers', 'Integer number of buffers (must be even). Maximum \
                 lag step to compute in each generation of downsampling.')
@describe_input('number_of_levels', 'Integer number defining how many generations of \
                 downsampling to perform, i.e., the depth of the binomial tree \
                 of averaged frames')
@output_names('g2', 'tau', 'images', 'labels')
@describe_output('g2', 'Normalized g2 data array with shape = (len(lag_steps), num_rois)')
@describe_output('tau', 'array describing tau (lag steps)')
@visible('images', False)
@visible('labels', False)
@visible('rois', False)
@visible('image_item', False)
@intent(PlotIntent,
        match_key='1-time Correlation',
        name='g2',
        xLogMode=True,
        labels={"bottom": "𝜏", "left": "g₂"},
        output_map={'x': 'tau', 'y': 'g2'},
        mixins=["ToggleSymbols"])
def one_time_correlation(images: np.ndarray,
                         labels: np.ndarray = None,
                         rois: Iterable[pg.ROI] = None,
                         image_item: pg.ImageItem = None,
                         num_bufs: int = 16,
                         num_levels: int = 8,
                         intensity_drift_correction: bool = True) -> Tuple[da.array, da.array, da.array, np.ndarray]:
    if images.ndim < 3:
        raise ValueError(f"Cannot compute correlation on data with {images.ndim} dimensions.")

    # if labels array was not passed in, it must be generated; trimming will occur here for memory conservation
    if labels is None:
        labels = get_label_array(images, rois=rois, image_item=image_item)
        if labels.max() == 0:
            msg.notifyMessage("Please add an ROI over which to calculate one-time correlation.")
            raise ValueError("Please add an ROI over which to calculate one-time correlation.")

        # Trim the image based on labels, and resolve to memory
        si, se = np.where(np.flipud(labels))
        # trimmed_images = np.asarray(images[:, si.min():si.max() + 1, se.min():se.max() + 1])
        trimmed_images = np.asarray([image[si.min():si.max() + 1, se.min():se.max() + 1] for image in images])
        trimmed_labels = np.asarray(np.flipud(labels)[si.min():si.max() + 1, se.min():se.max() + 1])


    # If a labels array is passed in, no trimming is done; autocorr should read each frame lazy-like
    else:
        trimmed_images = images
        trimmed_labels = labels

    # trimmed_images[trimmed_images <= 0] = np.NaN   # may be necessary to mask values

    if intensity_drift_correction:
        trimmed_images = trimmed_images / np.mean(trimmed_images, axis=(1, 2))[:, None, None]

    trimmed_images -= np.min(trimmed_images, axis=0)

    g2, tau = corr.multi_tau_auto_corr(num_levels, num_bufs,
                                       trimmed_labels.astype(np.uint8),
                                       trimmed_images)
    g2 = g2[1:].squeeze()
    # FIXME: is it required to trim the 0th value off the tau and g2 arrays?
    return g2.T, tau[1:], images, labels



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from xicam.plugins import QWidgetPlugin
from xicam.core import threads
from qtpy.QtWidgets import QVBoxLayout, QComboBox, QPushButton
from collections import OrderedDict

from astropy.modeling import fitting
from .factory import XicamSASModel
from .loader import load_models
import numpy as np

from pyqtgraph.parametertree import ParameterTree


class SASModelsWidget(QWidgetPlugin):
    '''
        documentation:
    '''
    name = 'SASModels'

    fitters = OrderedDict({
        'LinearLSQFitter': fitting.LinearLSQFitter,
        'LevMarLSQFitter': fitting.LevMarLSQFitter,
        'SLSQPLSQFitter': fitting.SLSQPLSQFitter
    })

    def __init__(self, *args, **kwargs):
        self.models_tree = load_models()
        self.categories = self.models_tree.keys()
        super().__init__(self, *args, *kwargs)

        # verticle layout
        vlayout = QVBoxLayout()

        # add a dropdown list of fitting routines
        self.fitterbox = QComboBox()
        self.fitterox.addItems(list(self.fitters.keys()))
        vlayout.addWidget(self.fitterbox)

        # add a dropdown list of categories
        self.catbox = QComboBox()
        self.catbox.addItems(self.categories)
        vlayout.addWidget(self.catbox)

        # add list of models in category
        cat = self.catbox.currentText()
        self.modelsbox = QComboBox()
        self.modelsbox.addItems(list(self.models_tree[cat]))
        vlayout.addWidget(self.modelsbox)

        # add parameter tree
        modelname = self.modelsbox.currentText()
        parameters = self.models_tree[cat][modelname]['params']
        param_tree = ParameterTree(showTop=False)
        param_tree.addParameters(self.parameters)
        vlayout.addWidget(param_tree)
        self.fittable = XicamSASModel(modelname, parameters)

        # add fit-button
        fit_button = QPushButton('Fit')
        fit_button.setToolTip('Fit model to the data')
        vlayout.addWidget(fit_button)
        fit_button.clicked.connect(self.run)

    def update_model(self):
        cat = self.catbox.currentText()
        modelname = self.modelsbox.currentText()
        parameters = self.models_tree[cat][modelname]['params']
        self.fittable = XicamSASModel(modelname, parameters)
        for p in parameters:
            if not p.name() in self.fittable.param_names:
                raise KeyError
            # set fixed if true
            self.fittable.fixed[p.name()] = p.child('Fixed').value()
            # set bounds if available
            if p.child('Bounded').value():
                bounds = (p.child('Bounded').child('Lower').value(),
                          p.child('Bounded').child('Upper').value())
                self.fittable.bounds[p.name()] = bounds
            else:
                self.fittable.bounds[p.name()] = (-np.inf, np.inf)

    def update(self, t):
        self.opt = t
        return t

    def run(self, q, I, callback_slot=None):
        if callback_slot is None: callback_slot = lambda t: None
        self.update_model()
        key = self.fitterbox.currentText()
        fitting_method = self.fitters[key]()
        thread = threads.QThreadFuture(fitting_method,
                                       self.fittable,
                                       q,
                                       I,
                                       callback_slot=lambda t: callback_slot(self.update(t)))
        thread.start()
        return thread



from databroker.core import BlueskyRun

from xicam.core.data import ProjectionException
from xicam.SAXS.stages import CorrelationStage, CalibrateGUIPlugin, ReduceGUIPlugin


class SAXSPlugin(CorrelationStage, ReduceGUIPlugin, CalibrateGUIPlugin):  # ...):
    name = 'SAXS'



DATA_PROJECTION_KEY = "/entry/instrument/detector/data"
ENERGY_PROJECTION_KEY = "/entry/instrument/monochromator/wavelength"
BEAM_CENTER_X_PROJECTION_KEY = "/entry/instrument/detector/beam_center_x"  # NOTE: These are lengths, not pixel pos's
BEAM_CENTER_Y_PROJECTION_KEY = "/entry/instrument/detector/beam_center_y"
X_PIXEL_SIZE_PROJECTION_KEY = "/entry/instrument/detector/x_pixel_size"  # NOTE: These might have to be resolved with
Y_PIXEL_SIZE_PROJECTION_KEY = "/entry/instrument/detector/y_pixel_size"  # the detector class
DISTANCE_PROJECTION_KEY = "/entry/instrument/detector/distance"
AZIMUTHAL_ANGLE_PROJECTION_KEY = "/entry/instrument/detector/azimuthal_angle"
INCIDENCE_ANGLE_PROJECTION_KEY = "/entry/instrument/sample/incidence_angle"
DETECTOR_TRANSLATION_X_PROJECTION_KEY = "/entry/instrument/detector/translate_x"
DETECTOR_TRANSLATION_Y_PROJECTION_KEY = "/entry/instrument/detector/translate_y"



from enum import Enum

import numpy as np
from xicam.plugins.operationplugin import operation, output_names, display_name, describe_input, describe_output, \
    categories, units
from pyFAI import calibrant
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
from scipy import signal

# TODO: use Multigeometry
# Create Enum containing a dict of calibrations (ordered according to keys)
calibrants = Enum("Calibrant", dict(sorted(calibrant.ALL_CALIBRANTS.all.items())))


@operation
@output_names('azimuthal_integrator')
@display_name('Sample-Detector Distance estimation (Naive)')
@describe_input('data', 'Calibrant frame image data')
@describe_input('mask', "Array (same size as image) with 1 for masked pixels, and 0 for valid pixels")
@describe_input('calibrant', "Calibrant standard record")
@describe_input('azimuthal_integrator', "Azimuthal integrator; the SDD will be modified in-place")
@describe_input('npts', "Resolution in q of the azimuthal integration used for ring detection")
@describe_input('wavelength_override', "If provided, override the wavelength provided by the input azimuthal_integrator")
@units('wavelength_override', 'm')
@describe_output('azimuthal_integrator', "Azimuthal integrator with the SDD be modified in-place")
@categories(('Scattering', 'Calibration'))
def naive_sdd(data: np.ndarray,
              azimuthal_integrator: AzimuthalIntegrator,
              calibrant: calibrants = calibrant.get_calibrant("AgBh"),
              mask: np.ndarray=None,
              wavelength_override: float = None,
              npts: int = 2000) -> AzimuthalIntegrator:
    kwargs = {}
    if mask is not None:
        kwargs['mask'] = mask

    # TODO: add a type that is a special parameter-tree item allowing enable/disable the parameter in the gui
    if wavelength_override:
        azimuthal_integrator.set_wavelength(wavelength_override)

    # slice into the first index as long as there's higher dimensionality
    while len(data.shape) > 2:
        data = data[0]

    # Un-calibrated azimuthal integration
    r, radialprofile = azimuthal_integrator.integrate1d(np.asarray(data), npts, unit='r_mm', **kwargs)

    # find peaks
    peaks = np.array(find_peaks(np.arange(len(radialprofile)), radialprofile)).T

    # get best peak
    bestpeak = None
    for peak in peaks:
        if peak[0] > 15 and not np.isinf(peak[1]):  ####This thresholds the minimum sdd which is acceptable
            bestpeak = peak[0]
            # print peak
            break

    # identify order of selected peak
    best_order = (0, 0)

    for i in range(1, 6):
        peak_ratios = ((peaks[:, 0] / (np.arange(len(peaks)))) / (bestpeak / (i + 1)))
        order = np.sum(np.logical_and(peak_ratios < 1.1, 0.9 < peak_ratios))
        if order > best_order[0]:
            best_order = (order, i)

    calibrant1stpeak = calibrant.dSpacing[best_order[1]]

    # Calculate sample to detector distance for lowest q peak

    tth = 2 * np.arcsin(0.5 * azimuthal_integrator.wavelength / calibrant1stpeak / 1.e-10)
    tantth = np.tan(tth)
    sdd = r[int(round(bestpeak))] / tantth

    # set sdd back on azimuthal integrator
    fit2dcal = azimuthal_integrator.getFit2D()
    fit2dcal['directDist'] = sdd
    azimuthal_integrator.setFit2D(**fit2dcal)

    return azimuthal_integrator


def find_peaks(x, y):
    # Default peak detection parameters
    wavelet = signal.ricker  # wavelet of choice
    widths = np.arange(1, 20)  # range of widths of the ricker wavelet to search/evaluate
    max_distances = widths / 8.  # ridgeline connectivity threshold; smaller values gives more peaks; larger values considers overlapping peaks as one
    gap_thresh = 4  # threshold number of rows for ridgeline connectivity; smaller values gives more peaks
    min_length = 3  # minimum ridgeline length; smaller values gives more peaks
    min_snr = 2  # Minimum SNR
    noise_perc = 10  # percentile of points below which to consider noise
    h = 3  # number of points skipped in finite differences
    truncationlow = 10  # low q truncation for zeros
    truncationhigh = 50

    peaks = signal.find_peaks_cwt(y, widths, wavelet, max_distances, gap_thresh, min_length, min_snr, noise_perc)
    peaks = peaks[1:]
    return list(np.array(np.vstack([x[peaks], y[peaks], peaks])))



from typing import Tuple

from xicam.core.intents import PlotIntent
from xicam.plugins.operationplugin import operation, output_names, display_name, describe_input, describe_output, \
    categories, intent
import numpy as np
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator


@operation
@display_name("X Integration")
@output_names("q_x", "I")
@describe_input("azimuthal_integrator", 'A PyFAI.AzimuthalIntegrator object')
@describe_input("data", '2d array representing intensity for each pixel')
@describe_input("mask", 'Array (same size as image) with 1 for masked pixels, and 0 for valid pixels')
@describe_input('dark', 'Dark frame image')
@describe_input('flat', 'Flat field image')
@describe_input('normalization_factor', 'Value of a normalization monitor')
@describe_output('q_x', "q_x bin center positions")
@categories(('Scattering', 'Integration'))
@intent(PlotIntent, name='X Integration', output_map={'x': 'q_x', 'y': 'I'}, labels={'bottom': 'q_x', 'left': 'I'})
@describe_output('I', 'Binned/pixel-split integrated intensity')
def x_integrate(azimuthal_integrator: AzimuthalIntegrator,
                data: np.ndarray,
                mask: np.ndarray = None,
                dark: np.ndarray = None,
                flat: np.ndarray = None,
                normalization_factor: float = 1) -> Tuple[np.ndarray, np.ndarray]:
    if data.ndim == 2:
        data = [data]
    if dark is None: dark = np.zeros_like(data[0])
    if flat is None: flat = np.ones_like(data[0])
    if mask is None: mask = np.zeros_like(data[0])
    I = []
    q_x = []
    for frame in data:
        frame = np.asarray(frame)
        I_i = np.sum((frame - dark) * np.average(flat - dark) / (flat - dark) / normalization_factor * np.logical_not(mask),
                     axis=-2)
        I.append(I_i)
        centerx = azimuthal_integrator.getFit2D()['centerX']
        centerz = azimuthal_integrator.getFit2D()['centerY']
        q_x_i = azimuthal_integrator.qFunction(np.array([centerz] * frame.shape[-1]),
                                               np.arange(0, frame.shape[-1])) / 10.
        q_x_i[np.arange(0, frame.shape[-1]) < centerx] *= -1.
        q_x.append(q_x_i)

        # TODO: support dynamic q
    return np.asarray(q_x_i), np.asarray(I)



from xicam.plugins import ProcessingPlugin, Input, Output, InOut
from pyFAI.detectors import Detector
import numpy as np
from pyqtgraph import ROI
from pyqtgraph.parametertree import parameterTypes
from typing import List, Tuple
from matplotlib.path import Path


class VerticalROI(ProcessingPlugin):
    name = 'Vertical ROI'

    ai = Input(
        description='PyFAI detector instance; the geometry of the detector''s inactive area will be masked.',
        type=Detector)
    polygon = Input(description='Polygon shape to mask (interior is masked)', type=List[Tuple[float, float]])
    mask = InOut(description='Mask array (1 is masked).', type=np.ndarray)

    def evaluate(self):
        if self.polygon.value is not None:
            # create path
            path = Path(np.vstack([self.polygon.value, self.polygon.value[-1]]),
                        [Path.MOVETO] + [Path.LINETO] * (len(self.polygon.value) - 1) + [Path.CLOSEPOLY])

            # create a grid
            ny, nx = self.ai.value.detector.shape
            x = np.linspace(0, nx, nx)
            y = np.linspace(0, ny, ny)
            xgrid, ygrid = np.meshgrid(x, y)
            pixel_coordinates = np.c_[xgrid.ravel(), ygrid.ravel()]

            # find points within path
            self.mask.value = np.logical_or(self.mask.value,
                                            np.flipud(path.contains_points(pixel_coordinates).reshape(ny, nx)))

    @property
    def parameter(self):
        if not (hasattr(self, '_param') and self._param):
            instructions = parameterTypes.TextParameter(name='Instructions',
                                                        value='Use the mouse to draw a mask. Click [Finish Mask] below when complete, or [Clear Selection] to start over.',
                                                        readonly=True)
            clearmask = parameterTypes.ActionParameter(name='Clear Selection')
            finishmask = parameterTypes.ActionParameter(name='Finish Mask')

            children = [instructions, clearmask, finishmask]
            self._param = parameterTypes.GroupParameter(name='Polygon Mask', children=children)
        return self._param



from enum import Enum
import numpy as np
from pyFAI.detectors import ALL_DETECTORS, Detector
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
from xicam.plugins.operationplugin import describe_input, describe_output, operation, output_names, display_name, units


def get_name(detector):
    if detector.aliases:
        name = detector.aliases[0]
    else:
        name = detector.__class__.__name__
    return name


# Remove 'detector' and sort by name
_detectors = {get_name(detector): detector
              for detector in ALL_DETECTORS.values()
              if detector is not Detector}
_sorted_detectors = dict(sorted(_detectors.items()))
DetectorEnum = Enum('Detector', _sorted_detectors)


@operation
@output_names()
@units("distance", "mm")
@units("center_x", "mm")
@units("center_y", "mm")
@units("tilt", "degrees")
@units("tilt_plane_rotation", "degrees")
@describe_input("distance", "Distance to detector center")
@describe_input("center_x", "")
@describe_input("center_y", "")
@describe_input("tilt", "")
@describe_input("tilt_plane_rotation", "")
def set_geometry(distance: float = 1,
                 center_x: float = 0,
                 center_y: float = 0,
                 tilt: float = 0.0,
                 tilt_plane_rotation: float = 0.0):
    ai = AzimuthalIntegrator()
    ai.setFit2D(distance, center_x, center_y, tilt, tilt_plane_rotation)
    return ai


# TODO: override parameter generation to disable detector selection when auto-mode is active
@operation
@display_name("Set Detector")
@output_names("azimuthal_integrator")
@describe_input("detector", "Detector to select in manual mode (automatic_mode = False).")
@describe_input("automatic_mode", "If True, automatically select a detector based on size heuristics; "
                                  "otherwise, uses the 'detector' input.")
@describe_input("binning_x", "Binning size of x dimension.")
@describe_input("binning_y", "Binning size of y dimension.")
@describe_input("azimuthal_integrator", "")
@describe_output("azimuthal_integrator", "")
def set_detector(detector: DetectorEnum = next(iter(DetectorEnum.__members__.values())).value,
                 automatic_mode: bool = True,
                 binning_x: int = 1,
                 binning_y: int = 1,
                 azimuthal_integrator: AzimuthalIntegrator = None,
                 data: np.ndarray = None) -> AzimuthalIntegrator:

    if automatic_mode:
        # Try to find best match based on the size
        selected_detector = guess_detector_by_shape(data.shape[-2:])

    else:
        selected_detector = detector()
        selected_detector.set_binning((binning_x, binning_y))

    ai = azimuthal_integrator or AzimuthalIntegrator()
    ai.detector = selected_detector
    return ai


def guess_detector_by_shape(shape):
    # for every detector known to pyFAI
    for name, detector in sorted(ALL_DETECTORS.items()):
        # if a shape limit is set
        if hasattr(detector, 'MAX_SHAPE'):
            if detector.MAX_SHAPE == shape:
                return detector()
        if hasattr(detector, 'BINNED_PIXEL_SIZE'):
            for binning in detector.BINNED_PIXEL_SIZE.keys():
                if shape == tuple(np.array(detector.MAX_SHAPE) / binning): # possibly needs to be reversed [::-1]
                    detector = detector()
                    detector.set_binning(binning)
                    return detector
    return None



from functools import partial
from typing import Dict, List

import numpy as np
from databroker.core import BlueskyRun
from databroker.in_memory import BlueskyInMemoryCatalog
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDialog, QListWidget, QHBoxLayout, QPushButton, QDialogButtonBox, \
    QVBoxLayout
from xicam.SAXS.workflows.roi import ROIWorkflow
from xicam.core import threads

from xicam.core.execution.workflow import ingest_result_set, project_intents
from xicam.core.intents import ROIIntent
from xicam.core.threads import invoke_in_main_thread
from xicam.gui.actions import Action
from xicam.gui.canvases import XicamIntentCanvas
from xicam.gui.models.treemodel import EnsembleModel
from xicam.gui.plugins.ensembleguiplugin import EnsembleGUIPlugin
from xicam.gui.widgets import PreviewWidget
from xicam.gui.widgets.linearworkfloweditor import WorkflowEditor
from xicam.plugins import GUILayout, manager as plugin_manager

from xicam.SAXS.calibration.workflows import SimulateWorkflow, CalibrationWorkflow
from xicam.SAXS.intents import SAXSImageIntent, GISAXSImageIntent
from xicam.SAXS.masking.workflows import MaskingWorkflow
from xicam.SAXS.operations.workflows import DisplayWorkflow, ReduceWorkflow
from xicam.SAXS.projectors.edf import project_NXsas
from xicam.SAXS.projectors.nxcansas import project_nxcanSAS
from xicam.SAXS.widgets.SAXSViewerPlugin import QLabel
from xicam.SAXS.workflows.xpcs import OneTime, TwoTime


# FIXME: the old way used TabWidget.currentWidget with XPCSToolBar...
# - previous: view was a tab view with the SAXSReductionViewer mixin as its widget
# - how can we adapt this to StackedCanvasView / CanvasView?
# # SAXS GUI Plugin mixin can use shared components
# class SAXSGUIPlugin(CorrelationGUIPlugin, SAXSReductionGUIPlugin)


class BaseSAXSGUIPlugin(EnsembleGUIPlugin):
    name = "SAXS"
    # Re-implement abstract methods
    @property
    def exposedvars(self) -> Dict:
        pass

    def currentheader(self) -> Dict:
        pass

    def __init__(self):
        super(BaseSAXSGUIPlugin, self).__init__()

        # Add in appropriate projectors here
        # Make sure project_intents is last (since it is generic case)
        # FIXME: refactor separation of SAXS v. XPCS
        self._projectors.extend([project_NXsas, project_nxcanSAS, project_intents])

        self.maskingworkflow = MaskingWorkflow()
        self.roiworkflow = ROIWorkflow()
        self.simulateworkflow = SimulateWorkflow()
        self.displayworkflow = DisplayWorkflow()
        self.reduceworkflow = ReduceWorkflow()

    def get_active_images(self, _):
        intent_indexes = [self.intents_model.index(row, 0) for row in range(self.intents_model.rowCount())]
        intents = {index: index.internalPointer()
                   for index in intent_indexes}
        image_indexes = list(filter(lambda index: isinstance(intents[index], SAXSImageIntent), intents.keys()))

        # FIXME: handle multiple images
        if len(image_indexes) > 1:
            image_names = [img.parent().data(Qt.DisplayRole) for img in image_indexes]
            raise ValueError("Running multiple images in this workflow is currently unsupported.")
        elif len(image_indexes) == 0:
            raise ValueError("No images are selected; cannot run this workflow.")

        image_index = image_indexes[0]
        canvas = self.intents_model.data(image_index, self.intents_model.canvas_role)
        # Test with time-series
        kwargs = {'images': np.squeeze(intents[image_index].image),
                  'image_item': canvas.canvas_widget.imageItem,
                  'geometry': intents[image_index].geometry}

        if 'darks' in intents[image_index].kwargs:
            kwargs['darks'] = np.squeeze(intents[image_index].kwargs['darks'])

        # Provide incidence angle and transmission mode based on SAXS v. GISAXS image intent type
        # TODO: if we support multiple image_indexes, will need to handle appropriately
        if isinstance(intents[image_index], GISAXSImageIntent):
            kwargs['transmission_mode'] = 'reflection'
            kwargs['incidence_angle'] = image_index.internalPointer().incidence_angle

        # Return the visualized (checked) rois as well
        roi_intents = list(filter(lambda intent: isinstance(intent, ROIIntent), intents.values()))
        rois = list(map(lambda intent: intent.roi, roi_intents))

        kwargs['rois'] = rois
        return kwargs


class ReduceGUIPlugin(BaseSAXSGUIPlugin):
    name = "Reduce"

    def __init__(self):
        super(ReduceGUIPlugin, self).__init__()

        self.reduce_panel = WorkflowEditor(self.reduceworkflow,
                                           kwargs_callable=self.begin_reduce,
                                           wavelength_override=0.124e-9,
                                           callback_slot=self.append_reduced)

        self.mask_panel = WorkflowEditor(self.maskingworkflow)
        self.display_panel = WorkflowEditor(self.displayworkflow)

        self.reduce_layout = GUILayout(self.canvases_view,
                                       right=self.ensemble_view,
                                       rightbottom=self.reduce_panel)
        stages = {'Reduce': self.reduce_layout}

        self.stages.update(**stages)

    def begin_masking(self, _, **kwargs):
        results = self.maskingworkflow.execute_synchronous(**kwargs)
        return list(results)[-1]

    def begin_roi(self, _, **kwargs):
        results = self.roiworkflow.execute_synchronous(**kwargs)
        return list(results)[-1]

    def begin_reduce(self, _):
        # # get catalogs from active ensemble
        # active_ensemble = self.ensemble_model.activeEnsemble
        # if not active_ensemble:
        #     raise RuntimeError("Unable to reduce since there is no data currently loaded. "
        #                        "Try opening data from the data browser on the left, "
        #                        "then retry running the reduction workflow.")
        #
        # # [] handles case where there is an active ensemble but no catalogs under it
        # active_catalogs = self.ensemble_model.catalogs() or []
        #
        # if not active_catalogs:
        #     raise RuntimeError("There are no catalogs in the active ensemble "
        #                        f'"{active_ensemble.data(Qt.DisplayRole)}". '
        #                        "Unable to reduce.")
        #
        # class ReduceDialog(QDialog):
        #     """Dialog for calibrating images.
        #
        #     User can select from a list of catalogs (pulled from the active ensemble),
        #     preview, and calibrate the image data.
        #     """
        #     def __init__(self, catalogs: List[BlueskyRun], parent=None, window_flags=Qt.WindowFlags()):
        #         super(ReduceDialog, self).__init__(parent, window_flags)
        #
        #         self.preview_widget = PreviewWidget()
        #
        #         self._catalogs = catalogs
        #
        #         self.catalog_selector = QListWidget()
        #         self.catalog_selector.addItems(map(lambda catalog: catalog.name, self._catalogs))
        #         self.catalog_selector.currentRowChanged.connect(self._update_preview)
        #
        #         calibrate_button = QPushButton("&Reduce")
        #         calibrate_button.setDefault(True)
        #
        #         self.buttons = QDialogButtonBox(Qt.Horizontal)
        #         # Add calibration button that accepts the dialog (closes with 1 status)
        #         self.buttons.addButton(calibrate_button, QDialogButtonBox.AcceptRole)
        #         # Add a cancel button that will reject the dialog (closes with 0 status)
        #         self.buttons.addButton(QDialogButtonBox.Cancel)
        #
        #         self.buttons.rejected.connect(self.reject)
        #         self.buttons.accepted.connect(self.accept)
        #
        #         layout = QHBoxLayout()
        #         layout.addWidget(self.catalog_selector)
        #         layout.addWidget(self.preview_widget)
        #
        #         outer_layout = QVBoxLayout()
        #         outer_layout.addLayout(layout)
        #         outer_layout.addWidget(self.buttons)
        #         self.setLayout(outer_layout)
        #
        #     def _update_preview(self, row: int):
        #         self.preview_widget.preview_catalog(self._catalogs[row])
        #
        #     def get_catalog(self):
        #         return self._catalogs[self.catalog_selector.currentRow()]
        #
        # if not active_catalogs:
        #     raise RuntimeError("There are no catalogs in the active ensemble "
        #                        f'"{active_ensemble.data(Qt.DisplayRole)}". '
        #                        "Unable to calibrate.")
        #
        # dialog = ReduceDialog(active_catalogs)
        # accepted = dialog.exec_()
        #
        # # Only calibrate if the dialog was accepted via the calibrate button
        # if not accepted == QDialog.Accepted:
        #     raise ValueError('Cancelled by user.')
        #
        # catalog = dialog.get_catalog()
        #
        # # TODO: better user feedback that there are no catalogs? (is that possible?)
        # if not catalog:
        #     raise ValueError('No catalog selected.')
        #
        # # find the saxsimageintent in this catalog
        # intents = self.ensemble_model.intents(catalog)
        #
        # image_intent = next(iter(filter(lambda intent: isinstance(intent, SAXSImageIntent), intents)))
        # data = image_intent.image
        #
        # kwargs = {"data": data}
        #
        # geometry = getattr(image_intent, 'geometry', None)
        # if geometry:
        #     kwargs['azimuthal_integrator'] = geometry
        #
        # masking_kwargs = self.begin_masking(None, **kwargs)
        #
        # kwargs.update(masking_kwargs)

        kwargs = self.get_active_images(None)
        kwargs['data'] = kwargs['images']
        kwargs['azimuthal_integrator'] = kwargs.get('geometry', None)

        masking_kwargs = self.begin_masking(None, **kwargs)
        kwargs.update(masking_kwargs)
        roi_kwargs = self.begin_roi(None, **kwargs)
        kwargs.update(roi_kwargs)

        kwargs['mask'] = kwargs['roi_mask']

        return kwargs

    def append_reduced(self, *results):
        document = list(ingest_result_set(self.reduceworkflow, results))
        # TODO: do we want to keep in memory catalog or write to attached databroker?
        # FIXME: use better bluesky_live design instead of upserting directly
        catalog = BlueskyInMemoryCatalog()
        catalog.upsert(document[0][1], document[-1][1], partial(iter, document), [], {})
        self.appendCatalog(catalog[-1])


class CalibrateGUIPlugin(BaseSAXSGUIPlugin):
    name = "Calibrate"

    def __init__(self):
        super(CalibrateGUIPlugin, self).__init__()

        self.calibration_workflow = CalibrationWorkflow()
        self.calibration_panel = WorkflowEditor(self.calibration_workflow,
                                                kwargs_callable=self.begin_calibrate,
                                                wavelength_override=0.124e-9,
                                                callback_slot=self.set_calibration)

        self.calibrate_layout = GUILayout(self.canvases_view,
                                          right=self.ensemble_view,
                                          rightbottom=self.calibration_panel)
        stages = {'Calibrate': self.calibrate_layout}

        self.stages.update(**stages)

    def begin_calibrate(self, _):
        # # get catalogs from active ensemble
        # active_ensemble = self.ensemble_model.activeEnsemble
        # if not active_ensemble:
        #     raise RuntimeError("Unable to calibrate since there is no data currently loaded. "
        #                        "Try opening data from the data browser on the left, "
        #                        "then retry running the calibration workflow.")
        #
        # # [] handles case where there is an active ensemble but no catalogs under it
        # active_catalogs = self.ensemble_model.catalogs() or []
        #
        # class CalibrationDialog(QDialog):
        #     """Dialog for calibrating images.
        #
        #     User can select from a list of catalogs (pulled from the active ensemble),
        #     preview, and calibrate the image data.
        #     """
        #     def __init__(self, catalogs: List[BlueskyRun], parent=None, window_flags=Qt.WindowFlags()):
        #         super(CalibrationDialog, self).__init__(parent, window_flags)
        #
        #         self.preview_widget = PreviewWidget()
        #
        #         self._catalogs = catalogs
        #
        #         self.catalog_selector = QListWidget()
        #         self.catalog_selector.addItems(map(lambda catalog: catalog.name, self._catalogs))
        #         self.catalog_selector.currentRowChanged.connect(self._update_preview)
        #
        #         calibrate_button = QPushButton("&Calibrate")
        #         calibrate_button.setDefault(True)
        #
        #         self.buttons = QDialogButtonBox(Qt.Horizontal)
        #         # Add calibration button that accepts the dialog (closes with 1 status)
        #         self.buttons.addButton(calibrate_button, QDialogButtonBox.AcceptRole)
        #         # Add a cancel button that will reject the dialog (closes with 0 status)
        #         self.buttons.addButton(QDialogButtonBox.Cancel)
        #
        #         self.buttons.rejected.connect(self.reject)
        #         self.buttons.accepted.connect(self.accept)
        #
        #         layout = QHBoxLayout()
        #         layout.addWidget(self.catalog_selector)
        #         layout.addWidget(self.preview_widget)
        #
        #         outer_layout = QVBoxLayout()
        #         outer_layout.addLayout(layout)
        #         outer_layout.addWidget(self.buttons)
        #         self.setLayout(outer_layout)
        #
        #     def _update_preview(self, row: int):
        #         self.preview_widget.preview_catalog(self._catalogs[row])
        #
        #     def get_catalog(self):
        #         return self._catalogs[self.catalog_selector.currentRow()]
        #
        # if not active_catalogs:
        #     raise RuntimeError("There are no catalogs in the active ensemble "
        #                        f'"{active_ensemble.data(Qt.DisplayRole)}". '
        #                        "Unable to calibrate.")
        #
        # dialog = CalibrationDialog(active_catalogs)
        # accepted = dialog.exec_()
        #
        # # Only calibrate if the dialog was accepted via the calibrate button
        # if not accepted == QDialog.Accepted:
        #     raise ValueError('Cancelled by user.')
        #
        # catalog = dialog.get_catalog()
        #
        # # TODO: better user feedback that there are no catalogs? (is that possible?)
        # if not catalog:
        #     raise ValueError('No catalog selected.')
        #
        # # find the saxsimageintent in this catalog
        # intents = self.ensemble_model.intents(catalog)
        #
        # image_intent = next(iter(filter(lambda intent: isinstance(intent, SAXSImageIntent), intents)))
        # data = image_intent.image

        kwargs = self.get_active_images(None)
        kwargs['data'] = kwargs['images']
        kwargs['azimuthal_integrator'] = kwargs.get('geometry', None)

        return kwargs

    def set_calibration(self, results):
        # TODO: confirmation dialog of calibration (or dialog of error)
        print(results)
        ai = results['azimuthal_integrator']

        # Find all intents within active ensemble, and set their geometry
        saxs_image_intents = [intent
                              for intent in
                              self.ensemble_model.intents(self.ensemble_model.activeEnsemble)
                              if isinstance(intent, SAXSImageIntent)]

        def _set_geometry(intent):
            intent.geometry = ai

        _ = list(map(_set_geometry, saxs_image_intents))

        calibration_settings = plugin_manager.get_plugin_by_name('xicam.SAXS.calibration', 'SettingsPlugin')
        device_names = list(map(lambda intent: intent.device_name, saxs_image_intents))

        for device_name in device_names:
            if device_name:
                calibration_settings.setAI(ai, device_name)

        # drop all canvases and refresh
        invoke_in_main_thread(self.canvases_view.refresh)


class CompareGUIPlugin(BaseSAXSGUIPlugin):
    name = "Compare"

    def __init__(self):
        super(CompareGUIPlugin, self).__init__()

        self.comparemultiview = QLabel("...")

        stages = {'Compare': GUILayout(self.comparemultiview,
                                       right=self.ensemble_view)}
        self.stages.update(**stages)


class CorrelationStage(BaseSAXSGUIPlugin):
    """
    Xi-CAM stage for XPCS correlation viewing and analysis.

    Features:
    * Open supported files in the data browser on the left, a preview will be shown when single-clicking the file.
    * Open configured databrokers to access and load bluesky runs.
    * Interactive tree-like viewer on the right-hand widget, where you can check/uncheck items to visualize and run analysis on.
    * Define ROIs on an opened and checked image item using the toolbar above the center widget.
    * Configurable and runnable 1-Time and 2-Time workflows in the bottom right widget.
    """
    name = "Correlation"

    # TODO: This doesn't really need to be two separate stages...
    def __init__(self):
        super(CorrelationStage, self).__init__()

        workflows = {OneTime(): '1-Time Correlation', TwoTime(): '2-Time Correlation'}
        self.workflow = next(iter(workflows.keys()))
        self.workflow.auto_connect_all()

        # FIXME: modify WorkflowEditor.workflow setter to allow changing workflow, kwargs_callable, etc.
        self.workflow_editor = WorkflowEditor(self.workflow,
                                              kwargs_callable=self.get_active_images,
                                              callback_slot=self.workflow_finished,
                                              workflows=workflows)
        self.gui_layout_template["rightbottom"] = self.workflow_editor
        stages = {"Correlation": GUILayout(**self.gui_layout_template)}
        self.stages.update(stages)

        self._roi_added = False

    def workflow_finished(self, *results):
        document = list(ingest_result_set(self.workflow, results))
        # TODO: do we want to keep in memory catalog or write to attached databroker?
        # FIXME: use better bluesky_live design instead of upserting directly
        catalog = BlueskyInMemoryCatalog()
        catalog.upsert(document[0][1], document[-1][1], partial(iter, document), [], {})
        self.appendCatalog(catalog[-1])

    def process_action(self, action: Action, canvas: XicamIntentCanvas):
        if not action.isAccepted():
            # Create ROI Intent adjacent to visualized intent
            roi_intent = ROIIntent(name=str(action.roi), roi=action.roi, match_key=canvas._primary_intent.match_key)
            catalog = self.ensemble_model.tree.parent(canvas._primary_intent)
            self.ensemble_model.appendIntent(roi_intent, catalog)

            self.workflow.auto_connect_all()
            action.accept()
        super(CorrelationStage, self).process_action(action, canvas)



import numpy as np
from xicam.plugins import ProcessingPlugin, Input, InOut
from scipy.ndimage import morphology


class ThresholdMaskPlugin(ProcessingPlugin):
    name = 'Threshold Mask'

    data = Input(description='Frame image data',
                 type=np.ndarray)
    minimum = Input(description='Threshold floor',
                    type=int,
                    default=3)
    maximum = Input(description='Threshold ceiling',
                    type=int,
                    default=1e10)
    neighborhood = Input(description='Neighborhood size in pixels for morphological opening. Only clusters of this size'
                                     ' that fail the threshold are masked',
                         type=int,
                         default=2)
    mask = InOut(description='Thresholded mask (1 is masked)',
                 type=np.ndarray)

    def evaluate(self):
        mask = np.logical_or(self.data.value < self.minimum.value, self.data.value > self.maximum.value)

        y, x = np.ogrid[-self.neighborhood.value:self.neighborhood.value + 1,
               -self.neighborhood.value:self.neighborhood.value + 1]
        kernel = x ** 2 + y ** 2 <= self.neighborhood.value ** 2

        morphology.binary_opening(mask, kernel, output=mask)  # write-back to mask
        if self.mask.value is not None:
            mask = np.logical_or(mask, self.mask.value)  # .astype(np.int, copy=False)
        self.mask.value = mask



import time
from pathlib import Path
from typing import Dict

import event_model
import h5py

from xicam.core.msg import WARNING, notifyMessage, logError, logMessage
from xicam.plugins.datahandlerplugin import DataHandlerPlugin


class APSXPCS(DataHandlerPlugin):
    """
    Handles ingestion of APS XPCS .hdf files.

    Internally, these .hdf files will hold a reference to a .bin image file,
    which will be loaded as well.
    """
    name = 'APSXPCS'
    DEFAULT_EXTENTIONS = ['.hdf', '.h5']

    def __init__(self, path):
        super(APSXPCS, self).__init__()
        self.path = path

    def __call__(self, data_key='', slc=0, **kwargs):
        # TODO -- change to show image data once image data is being ingested
        h5 = h5py.File(self.path, 'r')
        if data_key == 'norm-0-g2':
            return h5['exchange'][data_key][:,:,slc].transpose()
        return h5['exchange'][data_key][slc]

    @classmethod
    def ingest(cls, paths):
        updated_doc = dict()
        # TODO -- update for multiple paths (pending dbheader interface)
        if len(paths) > 1:
            paths = [paths[0]]
            message = 'Opening multiple already-processed data sources is not yet supported. '
            message += f'Opening the first image, {paths[0]}...'
            notifyMessage(message, level=WARNING)
            print(f'PATHS: {paths}')
        for name, doc in cls._createDocument(paths):
            if name == 'start':
                updated_doc[name] = doc
                # TODO -- should 'sample_name' and 'paths' be something different?
                doc['sample_name'] = cls.title(paths)
                doc['paths'] = paths
            if name == 'descriptor':
                if updated_doc.get('descriptors'):
                    updated_doc['descriptors'].append(doc)
                else:
                    updated_doc['descriptors'] = [doc]
            if name == 'event':
                if updated_doc.get('events'):
                    updated_doc['events'].append(doc)
                else:
                    updated_doc['events'] = [doc]
            if name == 'stop':
                updated_doc[name] = doc

        return updated_doc

    @classmethod
    def title(cls, paths):
        """Returns the title of the start_doc sample_name"""
        # return the file basename w/out extension
        # TODO -- handle multiple paths
        return Path(paths[0]).resolve().stem

    @classmethod
    def _find_datasets(cls, dataset_shape_indexes: Dict, h5: h5py.File):
        """
        Finds the HDF5 datasets based on the HDF5 paths in `dataset_shape_indexes`, and determines and stores the number
        of records for each of these datasets.

        Parameters
        ----------
        dataset_shape_indexes
            Dict with 'h5group/h5dataset' as keys and the shape index to use when determining the number of records
            in the dataset.
        h5
            The h5 file object.

        Returns
        -------
        dict
            Dictionary of the h5 dataset keys with values:
                dataset - the HDF5 dataset object
                num_items - the number of items in each dataset record

        """
        dataset_sizes = dict()
        for key, shape_index in dataset_shape_indexes.items():
            group = key.split('/')[0]
            dataset = key.split('/')[1]
            try:
                dataset_sizes[dataset] = {
                    'dataset': h5[group][dataset],
                    'num_items': [h5[group][dataset].shape[shape_index]],
                }
            except KeyError as ex:
                logMessage(f"[{key}] not found while ingesting [{h5.filename}].")
        return dataset_sizes

    @classmethod
    def _createDocument(cls, paths):
        # TODO -- add frames after being able to read in bin images
        for path in paths:
            timestamp = time.time()

            run_bundle = event_model.compose_run()
            yield 'start', run_bundle.start_doc

            source = 'APS XPCS'  # TODO -- find embedded source info?
            frame_data_keys = {'frame': {'source': source, 'dtype': 'number', 'shape': []}}
            frame_stream_name = 'primary'
            frame_stream_bundle = run_bundle.compose_descriptor(data_keys=frame_data_keys,
                                                                name=frame_stream_name)
            yield 'descriptor', frame_stream_bundle.descriptor_doc

            h5 = h5py.File(path, 'r')

            # Define the keys we want to ingest here; they don't necessarily have to exist
            datasets_shape_indexes = {
                'exchange/norm-0-g2': 0,
                'exchange/norm-0-stderr': 0,
                'exchange/tau': -1,  # tau shape is transposed
                'exchange/g2avgFIT1': 0,
                'xpcs/dqlist': 0
            }

            # Grab the HDF5 datasets and their associated sizes, size being the number of items in the dataset array
            # (e.g. grab the 'norm-0-g2' dataset,
            datasets_and_sizes = cls._find_datasets(datasets_shape_indexes, h5)
            exchange_transpose = ['norm-0-g2', 'norm-0-stderr', 'g2avgFIT1']

            reduced_data_keys = dict()
            for dataset in datasets_and_sizes:
                try:
                    reduced_data_keys[dataset] = {
                        'source': source,
                        'dtype': 'number',
                        'shape': datasets_and_sizes[dataset]['num_items'],
                    }
                except Exception as ex:
                    logError(ex)

            result_stream_name = '1-Time'
            reduced_stream_bundle = run_bundle.compose_descriptor(data_keys=reduced_data_keys,
                                                                  name=result_stream_name)
            yield 'descriptor', reduced_stream_bundle.descriptor_doc

            frames = []
            # TODO -- use the processed data timestamp?
            for frame in frames:
                yield 'event', frame_stream_bundle.compose_event(data={'frame', frame},
                                                                 timestamps={'frame', timestamp})


            # Define datasets that are the same for all events (not zipped)
            # e.g. the same tau dataset is used for all of the events
            consistent_datasets = ['tau']
            # Grab the actual arrays from each dataset, transposing so that the first dimension of the array
            # represents the number of dataset items. (e.g. norm-0-g2 might be stored in the dataset as an array
            # of shape (50, 3), which represents 3 g2 curves, with each g2 curve containing 50 points; so, we want
            # to transpose that for zipping)
            array_data = dict()
            for name, dataset in datasets_and_sizes.items():
                if name not in consistent_datasets:
                    if name in exchange_transpose:
                        array_data[name] = dataset['dataset'][()].squeeze().T
                    else:
                        array_data[name] = dataset['dataset'][()].squeeze()

            # Zip according to the dataset arrays' first dimensions (grab each 'curve' from the array of curves)
            for zipped in zip(*array_data.values()):

                # Compose a new dictionary with the dataset keys and the zipped dataset item (e.g. 'curve')
                data = dict(zip(array_data.keys(), zipped))
                data['dqlist'] = QROI(data['dqlist'])
                for dataset_key in consistent_datasets:
                    data[dataset_key] = datasets_and_sizes[dataset_key]['dataset'][()].squeeze()
                timestamps = dict(zip(data.keys(), len(data.keys()) * [timestamp]))
                yield 'event', reduced_stream_bundle.compose_event(data=data, timestamps=timestamps)

            yield 'stop', run_bundle.compose_stop()


class QROI(object):
    """
    Temporary class to put the qlist into a repr-able object so all ROIs can be treated generically.
    """
    def __init__(self, q_value):
        self.q = q_value

    def __repr__(self):
        return f"q = {self.q: .3g}"



from collections import namedtuple

import fabio
import numpy as np
from pyFAI import detectors, calibrant
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
import pytest

from xicam.core import execution
from xicam.core.execution import localexecutor
from xicam.SAXS.calibration.workflows import FourierCalibrationWorkflow
from xicam.SAXS.workflows.xpcs import OneTime, TwoTime


execution.executor = localexecutor.LocalExecutor()


def test_FourierCalibrationWorkflow():
    workflow = FourierCalibrationWorkflow()

    data = fabio.open('/home/rp/data/YL1031/AGB_5S_USE_2_2m.edf').data
    ai = AzimuthalIntegrator()
    ai.set_wavelength(124e-12)
    ai.detector = detectors.Pilatus2M()
    c = calibrant.ALL_CALIBRANTS('AgBh')

    print(workflow.execute(None, data=data, ai=ai, calibrant=c, callback_slot=print))


FRAMES = 100
SHAPE = (10, 10)

@pytest.fixture
def images():
    return np.random.random((FRAMES, *SHAPE))

@pytest.fixture
def labels():
    return np.ones(SHAPE)


class TestOneTimeWorkflow:
    def test_no_darks(self, images, labels):
        workflow = TwoTime()
        workflow.execute_synchronous(images=images, labels=labels)
        # TODO should dark correction be required?

    def test_with_darks(self):
        workflow = TwoTime()
        workflow.execute_synchronous(images=images,
                                     labels=labels,
                                     darks=None,
                                     flats=None)


class TestTwoTimeWorkflow:
    def test_no_darks(self, images, labels):
        workflow = TwoTime()
        result = workflow.execute_synchronous(images=images,
                                              labels=labels)
        print(result)

    def test_with_darks(self):
        ...


import numpy as np
from scipy import signal
from xicam.plugins.operationplugin import operation, output_names, display_name, describe_input, describe_output, \
    categories
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
from typing import Tuple


@operation
@output_names('beam_center', 'azimuthal_integrator')
@display_name('Fourier Autocorrelation')
@describe_input('data', "SAXS/WAXS calibrant image data")
@describe_input('azimuthal_integrator', "The `AzimuthalIntegrator` to modify; center will be modified in place")
@describe_output('beam_center', "Approximated position of the direct beam center")
@describe_output('azimuthal_integrator',
                 "The input `AzimuthalIntegrator` with its center shifted to the estimated center. If not provided, then `None`.")
@categories(('Scattering', 'Calibration'))
def fourier_autocorrelation(data: np.ndarray, azimuthal_integrator: AzimuthalIntegrator = None) -> \
        Tuple[Tuple[float, float], AzimuthalIntegrator]:
    """
    Estimate beam center of a SAXS/WAXS image using fourier autocorrelation. The validity of this technique depends on
    having negligible camera tilt/rotation, and a significant fraction of at least one ring feature.

    Parameters
    ----------
    data: np.ndarray
        SAXS/WAXS calibrant image data
    azimuthal_integrator: Optional[AzimuthalIntegrator]
        The `AzimuthalIntegrator` to modify; center will be modified in place

    Returns
    -------
    Tuple[float]
        Approximated position of the direct beam center
    Union[AzimuthalIntegrator, None]
        The input `AzimuthalIntegrator` with its center shifted to the estimated center. If not provided, then `None`.

    """

    mask = azimuthal_integrator.detector.mask
    data = np.squeeze(np.asarray(data))
    if mask is not None and mask.shape == data.shape:
        data = data * (1 - mask)

    # slice into the first index as long as there's higher dimensionality
    while len(data.shape) > 2:
        data = data[0]

    con = signal.fftconvolve(data, data) / np.sqrt(
        signal.fftconvolve(np.ones_like(data), np.ones_like(data)))

    center = np.array(np.unravel_index(con.argmax(), con.shape)) / 2.
    fit2dparams = azimuthal_integrator.getFit2D()
    fit2dparams['centerX'] = center[1]
    fit2dparams['centerY'] = center[0]
    azimuthal_integrator.setFit2D(**fit2dparams)

    return center, azimuthal_integrator



from xicam.SAXS.masking.roi import roi_mask_plugin
from xicam.core.execution.workflow import Workflow
# from xicam.gui.widgets.ROI import LabelArrayProcessingPlugin
from xicam.plugins.operationplugin import OperationPlugin


# TODO -- move to more reusable area
class ROIWorkflow(Workflow):
    """
    Workflow for ROIs.

    Initializes with a LabelArray operation.

    The expected result output will be through the `label_array` variable (defined in LabelArrayProcessingPlugin).
    """
    def __init__(self):
        super(ROIWorkflow, self).__init__(name="ROIWorkflow")
        self.add_operation(roi_mask_plugin())
        self.auto_connect_all()

    def prepend_operation(self, operation: OperationPlugin, auto_connect_all: bool = False):
        super(ROIWorkflow, self).insert_operation(0, operation)
        if auto_connect_all:
            self.auto_connect_all()




from typing import List

import numpy as np
import pyqtgraph as pg
from camsaxs.remesh_bbox import q_from_geometry

from xicam.SAXS.patches.pyFAI import AzimuthalIntegrator


def get_label_array(images: np.ndarray, rois: np.ndarray = None, image_item: pg.ImageItem = None, geometry: AzimuthalIntegrator = None) -> np.ndarray:
    while images.ndim > 2:
        images = images[0]

    if rois is None:
        return np.ones_like(images)
    # Create zeros label array to insert new labels into (if multiple ROIs)
    label_array = np.zeros(images.shape)
    roi_masks = []
    for roi in rois:
        # TODO: Should label array be astype(np.int) (instead of float)?
        if getattr(roi, 'is_Q_based', False):
            if not geometry:
                raise ValueError('No geometry provided.')
            label = roi.getLabelArray(images, image_item, geometry=geometry)
        else:
            label = roi.getLabelArray(images, image_item)

        # Store the boolean mask of each label array
        roi_mask = label.astype(np.bool)
        roi_masks.append(roi_mask)
        # Grab the current label array maximum value (so we can increment multiple labels accordingly)
        label_array_max = label_array.max()
        label = np.where(label > 0, label + label_array_max, label)
        # For single roi, our max will be 0 (since label_array is just np.zeros so far, hasn't been modified)
        if label_array_max == 0:
            label_array = label
        else:
            # FIXME right now, if labels overlap, label integers are being added together (into a new label value)
            # Adjust any currently non-masked areas with the new label
            label_array = np.where(label_array == 0, label, label_array)
            #
            label_array = np.where(label_array > 0,
                                   np.where(label > 0, label, label_array),
                                   label_array)
            # label_array = np.where(label_array > 0,
            #                        label or label_array,
            #                        label_array)

    return label_array.astype(np.int)


def average_q_from_labels(labels: np.ndarray,
                          geometry: AzimuthalIntegrator,
                          scattering_mode: str,
                          incidence_angle=None) -> List[float]:
    # q_h, q_v
    q = q_from_geometry(labels.shape, geometry, scattering_mode == 'reflection', incidence_angle or 0.0)

    # TODO: how can we allow choosing between these different q values (q_h, q_v, q_norm)
    # q magnitude
    q = np.linalg.norm(q, axis=2)
    average_qs = [np.average(q, weights=(labels == i)) for i in range(1, labels.max() + 1)]
    # TODO: return a dict mapping labels to qs?
    return average_qs



from xicam.core.intents import PlotIntent
from xicam.plugins.operationplugin import operation, output_names, display_name, describe_input, describe_output, \
    categories, intent
import numpy as np
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator


@operation
@display_name('SAXS q conversion')
@output_names('q_x', 'q_z')
@describe_input('integrator', 'A PyFAI.AzimuthalIntegrator object')
@describe_input('data', 'Frame image data')
@describe_output('q_x', 'q_x array with dimension of data')
@describe_output('q_y', 'q_y array with dimension of data')
@intent(PlotIntent, name='SAXS q conversion', output_names={'q_x': 'x', 'q_z': 'y'}, labels={'bottom': 'q_x',
                                                                                             'left': 'q_z'})
@categories(('Scattering', 'Transformations'))
def q_conversion_saxs(integrator: AzimuthalIntegrator,
                      data: np.ndarray) -> np.ndarray:
    chi = integrator.chiArray()
    twotheta = integrator.twoThetaArray()

    # TODO: Doble check what is chi = 0
    q_x = 2 * np.pi / integrator.getvalue('Wavelength') * np.sin(twotheta) * np.sin(chi)
    q_z = 2 * np.pi / integrator.getvalue('Wavelength') * np.sin(twotheta) * np.cos(chi)

    return q_x, q_z



from xicam.plugins import ProcessingPlugin, Input, Output, InOut
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
import fabio
import numpy as np


class FileMask(ProcessingPlugin):
    name = 'File Mask'

    mask = InOut(description='Mask array (1 is masked).',
                 type=np.ndarray)
    ai = Input(
        description='PyFAI azimuthal integrator instance; the geometry of the detector''s inactive area will be masked.',
        type=AzimuthalIntegrator)
    path = Input(description='File path to image mask file.', type=str, default='')

    def evaluate(self):
        if not self.path.value:
            return

        mask = fabio.open(self.path.value).data.astype(np.bool_)

        if not mask.shape == self.ai.value.detector.shape:
            raise IndexError('Mask file does not match detector shape.')

        if self.mask.value is None: self.mask.value = np.zeros(self.ai.value.detector.shape).astype(np.bool_)
        self.mask.value = np.logical_or(self.mask.value, mask)



import numpy as np
import pytest

from xicam.SAXS.operations.twotime import two_time_correlation


def test_twotime(self):
    # use dinesh's data
    ...

def test_twotime(self):
    data = np.random.random((100, 10, 10))
    labels = np.ones(data.shape[1:])
    print(data.shape)
    print(labels.shape)
    op = two_time_correlation()
    result = op(data=data, labels=labels)
    print(result)

    assert result == some_twotime_data



from xicam.plugins import ProcessingPlugin, Input, InOut
from scipy.ndimage import morphology
import numpy as np


class GrowMask(ProcessingPlugin):
    mask = InOut(description='Mask array (1 is masked).',
                 type=np.ndarray)
    size = Input(description='Distance to grow the mask (px); used as kernel size')

    def evaluate(self):
        y, x = np.ogrid[-self.size.value:self.size.value + 1,
               -self.size.value:self.size.value + 1]
        kernel = x ** 2 + y ** 2 <= self.size.value ** 2
        morphology.binary_dilation(self.mask.value, kernel, output=self.mask.value)  # write-back to mask



from xicam.core.execution.workflow import Workflow
from .qintegrate import q_integrate
from .chiintegrate import chi_integrate
from .xintegrate import x_integrate
from .zintegrate import z_integrate
from .cakeintegrate import cake_integration


class ReduceWorkflow(Workflow):
    def __init__(self):
        super(ReduceWorkflow, self).__init__('Reduce')

        self.qintegrate = q_integrate()
        self.chiintegrate = chi_integrate()
        self.xintegrate = x_integrate()
        self.zintegrate = z_integrate()

        self.add_operations(self.qintegrate, self.chiintegrate, self.xintegrate, self.zintegrate)
        self.auto_connect_all()


class DisplayWorkflow(Workflow):
    def __init__(self):
        super(DisplayWorkflow, self).__init__('Display')

        self.cake = cake_integration()
        self.add_operation(self.cake)
        self.auto_connect_all()



import numpy as np
from astropy.modeling import Fittable1DModel, Parameter, fitting, models
from xicam.SAXS.patches.pyFAI import AzimuthalIntegrator
from xicam.SAXS.utils import average_q_from_labels
from xicam.core import msg
from xicam.core.intents import PlotIntent, ScatterIntent
from xicam.plugins.operationplugin import operation, output_names, display_name, categories, intent, visible


@operation
@output_names('diffusion_fit', 'x', 'relaxation_rates', 'g2', 'tau', 'fit_curve')
@display_name('Diffusion Coefficient')
@categories(('XPCS',))
@intent(PlotIntent, 'Linear fit of Diffusion',
        output_map={'x': 'x', 'y': 'diffusion_fit'},
        labels={'bottom': 'q²', 'left': 'Γ'},
        match_key='diffusion_coefficient',
        mixins=["ToggleSymbols"])
@intent(ScatterIntent,
        'Diffusion Coefficients',
        output_map={'x': 'x', 'y': 'relaxation_rates'},
        labels={'bottom': 'q²', 'left': 'Γ'},
        match_key='diffusion_coefficient',
        mixins=["ToggleSymbols"])
# Make sure g2 and its fit intents are propagated to this end operation
@intent(PlotIntent,
        match_key='1-time Correlation',
        name='g2',
        yLog=True,
        labels={"bottom": "𝜏", "left": "g₂"},
        output_map={'x': 'tau', 'y': 'g2'},
        mixins=["ToggleSymbols"])
@intent(PlotIntent,
        canvas_name="1-time Correlation",
        match_key='1-time Correlation',
        name='g2 fit',
        labels={"bottom": "𝜏", "left": "g₂"},
        output_map={'x': 'tau', 'y': 'fit_curve'},
        mixins=["ToggleSymbols"])
@visible('labels', False)
@visible('relaxation_rates', False)
@visible('g2', False)
@visible('tau', False)
@visible('fit_curve', False)
@visible('geometry', False)
@visible('transmission_mode', False)
@visible('incidence_angle', False)
def diffusion_coefficient(relaxation_rates: np.ndarray,
                          labels: np.ndarray,
                          g2: np.ndarray,
                          tau: np.ndarray,
                          fit_curve: np.ndarray,
                          geometry: AzimuthalIntegrator = None,
                          transmission_mode: str = 'transmission',
                          incidence_angle: float = None,
                          ):
    # TODO: what should we do when we only get one relaxation rate (ie one roi / non-segmented roi)
    if geometry is None:
        msg.notifyMessage('Calibrate required for diffusion coefficients.')
        
        return np.array([0]), np.array([0]), relaxation_rates, g2, tau, fit_curve
        
    else:
        qs = np.asarray(average_q_from_labels(labels, geometry, transmission_mode, incidence_angle))

        x = qs ** 2
        # diffusion_values = relaxation_rates / x

        model = models.Linear1D()
        fitting_algorithm = fitting.LinearLSQFitter()

        fit = fitting_algorithm(model, x, relaxation_rates)

        return fit(x), x, relaxation_rates, g2, tau, fit_curve



from typing import Iterable

from xicam.SAXS.utils import get_label_array
from xicam.plugins.operationplugin import operation, output_names, display_name, describe_input, describe_output, \
    categories
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
import numpy as np
import pyqtgraph as pg


@operation
@display_name("ROI selection")
@describe_input('azimuthal_integrator', "The `AzimuthalIntegrator` to including a detector instance describing the camera hardware.")
@describe_output("mask", "The calculated mask derived from the detector's profile. (1 is masked)")
@output_names("roi_mask")
@categories(("Scattering", "Masking"))
def roi_mask_plugin(data: np.array = None,
                    rois: Iterable[pg.ROI] = None,
                    geometry: AzimuthalIntegrator = None,
                    image_item: pg.ImageItem = None,
                    mask: np.ndarray = None) -> np.ndarray:
    if rois:
        labels = get_label_array(data, rois=rois, image_item=image_item, geometry=geometry)
        return np.logical_or(mask, np.logical_not(labels))
    else:
        return mask



from xicam.plugins import ProcessingPlugin, Input, Output, InOut, PlotHint
import numpy as np
from astropy.modeling import fitting
from astropy.modeling import Fittable1DModel
from typing import Tuple
from enum import Enum
from xicam.plugins import manager as pluginmanager


class AstropyQSpectraFit(ProcessingPlugin):
    name = 'Q Fit (Astropy)'

    q = InOut(description='Q bin center positions',
              type=np.array)
    Iq = InOut(description='Q spectra bin intensities', type=np.array)
    model = Input(description='Fittable model class in the style of Astropy', type=Enum)
    domainmin = Input(description='Min bound on the domain of the input data', type=float)
    domainmax = Input(description='Max bound on the domain of the input data', type=float)
    fitter = Input(description='Fitting algorithm', default=fitting.LevMarLSQFitter(), type=Enum,
                   limits={'Linear LSQ': fitting.LinearLSQFitter(),
                           'Levenberg-Marquardt LSQ': fitting.LevMarLSQFitter(), 'SLSQP LSQ': fitting.SLSQPLSQFitter(),
                           'Simplex LSQ': fitting.SimplexLSQFitter()})

    fittedmodel = Output(description='A new model with the fitted parameters; behaves as parameterized function',
                         type=Fittable1DModel)
    fittedprofile = Output(
        description='The fitted profile from the evaluation of the resulting model over the input range.')

    hints = [PlotHint(q, Iq), PlotHint(q, fittedprofile)]

    modelvars = {}

    def __init__(self):
        super(AstropyQSpectraFit, self).__init__()
        self.model.limits = {plugin.name: plugin for plugin in
                             pluginmanager.get_plugins_of_type('Fittable1DModelPlugin')}
        self.model.value = list(self.model.limits.values())[0]

    @property
    def parameter(self):

        # clear cache in
        for input in self.modelvars:
            del self.__dict__[input]
        varcache = self.modelvars.copy()
        self.modelvars = {}
        self._inputs = None
        self._inverted_vars = None
        if hasattr(self, '_inverted_vars'): del self._inverted_vars

        for name in self.model.value.param_names:
            param = getattr(self.model.value, name)
            # TODO: CHECK NAMESPACE
            if name in varcache:
                input = varcache[name]
            else:
                input = InOut(name=name, default=param.default, limits=param.bounds, type=float, fixed=False,
                              fixable=True)
            setattr(self, name, input)
            self.modelvars[name] = input
        parameter = super(AstropyQSpectraFit, self).parameter
        parameter.child('model').sigValueChanged.connect(self.reset_parameter)
        return parameter

    def reset_parameter(self):
        # cache old parameter
        oldparam = self._param

        # empty it
        for child in oldparam.children():
            child.remove()

        # reset attribute so new parameter is generated
        self._param = None

        # add new children to old parameter
        for child in self.parameter.children():  # type: Parameter
            oldparam.addChild(child)

        # set old parameter to attribute
        self._param = oldparam

    def evaluate(self):
        if self.model.value is None or self.model.value == '----': return
        norange = self.domainmin.value == self.domainmax.value
        if self.domainmin.value is None and self.q.value is not None or norange:  # truncate the q and I arrays with limits
            self.domainmin.value = self.q.value.min()
        if self.domainmax.value is None and self.q.value is not None or norange:  # truncate the q and I arrays with limits
            self.domainmax.value = self.q.value.max()
        for name, input in self.modelvars.items():  # propogate user-defined values to the model
            getattr(self.model.value, name).value = input.value
            getattr(self.model.value, name).fixed = input.fixed
        filter = np.logical_and(self.domainmin.value <= self.q.value, self.q.value <= self.domainmax.value)
        q = self.q.value[filter]
        Iq = self.Iq.value[filter]
        self.fittedmodel.value = self.fitter.value(self.model.value, q, Iq)
        self.fittedprofile.value = self.fittedmodel.value(self.q.value)



from xicam.plugins.operationplugin import operation, output_names, display_name, describe_input, describe_output, \
    categories
import numpy as np


@operation
@display_name('Fourier Correlation')
@output_names('g2')
@describe_input('data', 'Array of two or more dimensions.')
@describe_input('labels', 'Labeled array of the same shape as the image stack.'
                          'Each ROI is represented by sequential integers starting at one.'
                          'For example, if you have four ROIs, they must be labeled'
                          '1, 2, 3, 4. Background is labeled as 0')
@describe_output('g2', 'Normalized correlation with shape = (len(lag_steps), num_rois)')
# TODO: intent
@categories('Scattering', 'Correlation')
def fourier_correlation(data: np.ndarray,
                        labels: np.ndarray) -> np.ndarray:
    data = np.array(data)
    mask = np.logical_not(1 == labels)

    x = np.asarray(data)
    N, NX, NY = x.shape
    x_mean = x.mean(axis=0)
    x_std = x.std(axis=0)
    x = x - x_mean

    # zero pad
    x = np.r_[x, np.zeros((N - 1, NX, NY))]
    s = np.fft.fft(x, axis=0)
    result = np.real(np.fft.ifft(s * s.conj(), axis=0))
    result = result[:N, :, :] / N / x_std
    g2 = np.ma.average(
        np.ma.masked_array(np.nan_to_num(result), mask=np.broadcast_to(mask, (result.shape[0],) + mask.shape)),
        axis=(1, 2)).data
    return g2



from typing import Callable, Union
from qtpy.QtWidgets import *
from qtpy.QtCore import *
from qtpy.QtGui import *
from xicam.plugins.widgetplugin import QWidgetPlugin
from xicam.gui.static import path
from xicam.core.execution.workflow import Workflow
from xicam.plugins import OperationPlugin
from xicam.gui.widgets.menuview import MenuView
from xicam.gui.widgets.ROI import ArcROI, LineROI, BetterPolyLineROI, BetterRectROI, SegmentedRectROI
from xicam.core import msg
from functools import partial
import pyqtgraph as pg


class SAXSToolbarBase(QToolBar):
    name = 'SAXSToolbar'
    sigPlotCache = Signal()
    sigDoWorkflow = Signal()
    sigDeviceChanged = Signal(str)

    def __init__(self, *args, **kwargs):
        super(SAXSToolbarBase, self).__init__(*args)

    def mkAction(self, iconpath: str = None, text=None, receiver=None, group=None, checkable=False, checked=False):
        actn = QAction(self)
        if iconpath: actn.setIcon(QIcon(QPixmap(str(path(iconpath)))))
        if text: actn.setText(text)
        if receiver: actn.triggered.connect(receiver)
        actn.setCheckable(checkable)
        if checked: actn.setChecked(checked)
        if group: actn.setActionGroup(group)
        return actn


class FieldSelector(SAXSToolbarBase):

    def __init__(self, headermodel: QStandardItemModel, selectionmodel: QItemSelectionModel, *args, **kwargs):
        self.headermodel = headermodel
        self.selectionmodel = selectionmodel
        self.headermodel.dataChanged.connect(self.updatedetectorcombobox)

        super(FieldSelector, self).__init__()

        self.addWidget(QLabel("Detector: "))
        self.detectorcombobox = QComboBox()
        self.detectorcombobox.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        self.addWidget(self.detectorcombobox)
        self.addSeparator()
        self.detectorcombobox.currentTextChanged.connect(self.sigDeviceChanged)

    def updatedetectorcombobox(self, start, end):
        if self.headermodel.rowCount():
            # TODO-- remove hard-coding of stream
            stream = "primary"
            item = self.headermodel.item(self.selectionmodel.currentIndex().row())
            # fields = getattr(item.data(Qt.UserRole), stream).to_dask().keys()
            catalog = item.data(Qt.UserRole)  # type: Catalog
            fields = [ technique["data_mapping"]["data_image"][1] for technique in catalog.metadata["techniques"] if technique["technique"] == "scattering" ]
            self.detectorcombobox.clear()
            self.detectorcombobox.addItems(fields)


class ModeSelector(SAXSToolbarBase):
    def __init__(self, *args, **kwargs):
        super(ModeSelector, self).__init__(*args, **kwargs)
        self.modegroup = QActionGroup(self)
        self.rawaction = self.mkAction('icons/raw.png', 'Raw', checkable=True, group=self.modegroup, checked=True)
        self.addAction(self.rawaction)
        self.cakeaction = self.mkAction('icons/cake.png', 'Cake (q/chi plot)', checkable=True, group=self.modegroup)
        self.addAction(self.cakeaction)
        self.remeshaction = self.mkAction('icons/remesh.png', 'Wrap Ewald Sphere', checkable=True, group=self.modegroup)
        self.addAction(self.remeshaction)
        self.addSeparator()

# TODO maybe toolbar is not the best solution here, instead having the buttonbox in the compare stage
# class ResultsModeSelector(SAXSToolbarBase):
#     def __init__(self, *args, **kwargs):
#         super(ResultsModeSelector, self).__init__(*args, **kwargs)
#         self.viewmodegroup = QActionGroup(self)
#         self.tabmode = self.mkAction(iconpath='icons/tabs.png', text='Tab View', checkable=True, group=self.viewmodegroup, checked=True)
#         self.addAction(self.tabmode)
#         self.gridmode = self.mkAction(iconpath='icons/grid.png', text='Grid View', checkable=True, group=self.viewmodegroup)
#         self.addAction(self.gridmode)
#         self.addSeparator()


class MultiPlot(SAXSToolbarBase):
    def __init__(self, *args, **kwargs):
        super(MultiPlot, self).__init__(*args, **kwargs)
        self.multiplot = QAction(self)
        self.multiplot.setIcon(QIcon(str(path('icons/multiplot.png'))))
        self.multiplot.setText('Plot Series')
        self.multiplot.setCheckable(True)
        self.multiplot.triggered.connect(self.sigDoWorkflow)
        self.addAction(self.multiplot)
        self.addSeparator()


class ROIs(SAXSToolbarBase):
    def __init__(self, *args, view: Union[Callable, pg.ImageView] = None, workflow=None, index=-1, **kwargs):
        super(ROIs, self).__init__(*args, **kwargs)
        self.workflow = workflow
        self.view = view
        self.index = index  # Where to insert the ROIs process into the workflow (default append)
        self._scale_factor = .33

        self.roi_button = QToolButton()
        self.roi_button.setText("Create ROI")
        self.roi_button.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
        self.roi_button.setPopupMode(QToolButton.InstantPopup)
        self.roi_menu = QMenu()
        self.roi_button.setMenu(self.roi_menu)
        # TODO -- disable button until we have loaded data

        self.arc_roi = self.mkAction('icons/roi_arc.png', 'Arc ROI', self.add_arc)
        self.roi_menu.addAction(self.arc_roi)
        # self.horizontal_roi = self.mkAction('icons/roi_horizontal.png', 'Horizontal ROI', self.add_horizontal)
        # self.roi_menu.addAction(self.horizontal_roi)
        self.line_roi = self.mkAction('icons/roi_line.png', 'Line ROI', self.add_line)
        self.roi_menu.addAction(self.line_roi)
        self.polygon_roi = self.mkAction('icons/roi_polygon.png', 'Polygon ROI', self.add_polygon)
        self.roi_menu.addAction(self.polygon_roi)
        self.rect_segmented_roi = self.mkAction('icons/roi_rect_segmented.png', 'Segmented Rectangular ROI',
                                                self.add_rect_segmented)
        self.roi_menu.addAction(self.rect_segmented_roi)
        self.rect_roi = self.mkAction('icons/roi_rect.png', 'Rectangular ROI', self.add_rect)
        self.roi_menu.addAction(self.rect_roi)
        # self.vertical_roi = self.mkAction('icons/roi_vertical.png', 'Vertical ROI', self.add_vertical)
        # self.roi_menu.addAction(self.vertical_roi)

        self.addWidget(self.roi_button)

        self.addSeparator()

    # TODO: scale roi's by inspecting self.view

    def _get_view(self):
        view = self.view
        if callable(view):
            view = view()
        return view

    def _scaled_size(self):
        view = self._get_view()
        if view:
            image_bound = view.imageItem.boundingRect()
            width = image_bound.width()
            height = image_bound.height()
            return width * self._scale_factor, height * self._scale_factor
        return -1, -1

    def _rect_origin(self):
        view = self._get_view()
        if view:
            image_bound = view.imageItem.boundingRect()
            width = image_bound.width()
            height = image_bound.height()
            origin_x = image_bound.x() + width / 2 - width / 2 * self._scale_factor
            origin_y = image_bound.y() + height / 2 - height / 2 * self._scale_factor
            return origin_x, origin_y
        return -1, -1

    def add_roi(self, roi):
        view = self._get_view()
        if view:
            view.getView().addItem(roi)
            self.workflow.insert_operation(self.index, roi.operation)
            # Remove the roi process from the workflow when the roi is removed
            # TODO -- should this be in BetterROI?
            roi.sigRemoveRequested.connect(lambda roi: self.workflow.remove_operation(roi.operation))
        else:
            msg.notifyMessage("Please open an image before creating an ROI.", level=msg.WARNING)

    def add_arc(self):
        self.add_roi(ArcROI(center=(0, 0), radius=.25))

    def add_horizontal(self):
        ...

    def add_line(self):
        image_bound = self.view().imageItem.boundingRect()
        width = image_bound.width()
        height = image_bound.height()
        x = image_bound.x() + width / 2 + width / 2 * self._scale_factor
        y = image_bound.y() + height / 2
        self.add_roi(LineROI(pos1=(self._rect_origin()[0], y), pos2=(x, y), width=self._scaled_size()[0]))

    def add_polygon(self):
        rect = QRectF(QPointF(*self._rect_origin()), QSizeF(*self._scaled_size()))
        points = [(point.x(), point.y()) for point in [rect.bottomLeft(),
                                                       rect.bottomRight(),
                                                       rect.topRight(),
                                                       rect.topLeft()]]
        self.add_roi(BetterPolyLineROI(points, closed=True))

    def add_rect(self):
        self.add_roi(BetterRectROI(pos=self._rect_origin(), size=self._scaled_size()))

    def add_rect_segmented(self):
        self.add_roi(SegmentedRectROI(pos=self._rect_origin(), size=self._scaled_size()))

    def add_vertical(self):
        ...


class SAXSToolbarRaw(FieldSelector):
    pass


class SAXSToolbarMask(FieldSelector):
    pass


class SAXSToolbarReduce(MultiPlot, ROIs, ModeSelector):
    def __init__(self, *args, **kwargs):
        super(SAXSToolbarReduce, self).__init__(*args, **kwargs)


# class  SAXSToolbarCompare(ResultsModeSelector):
#     pass




from qtpy.QtCore import Qt
from .SAXSToolbar import FieldSelector, ROIs, SAXSToolbarBase


# TODO -- move these to more reusable area
class Button(SAXSToolbarBase):
    def __init__(self, *args, button_icon_path=None, button_text="", receiver=None, **kwargs):
        super(Button, self).__init__(*args, **kwargs)
        action = self.mkAction(button_icon_path, button_text, receiver, **kwargs)
        self.addAction(action)
        if button_text:
            self.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)


class ProcessButton(Button):
    """Defines a button that can be used to start a process."""
    def __init__(self, *args, button_receiver=None, **kwargs):
        super(ProcessButton, self).__init__(*args,
                                            button_icon_path="icons/run.png",
                                            button_text="Process",
                                            receiver=button_receiver,
                                            **kwargs)


class XPCSToolBar(ROIs):
    pass


from xicam.SAXS.operations.select_detector import set_detector
from xicam.core.execution.workflow import Workflow
from .fourierautocorrelation import fourier_autocorrelation
from xicam.SAXS.calibration.simulatecalibrant import simulate_calibrant
from .naivesdd import naive_sdd


class CalibrationWorkflow(Workflow):
    def __init__(self):

        super(CalibrationWorkflow, self).__init__('Calibration')

        set_detector_op = set_detector()
        autocor_op = fourier_autocorrelation()
        naive_sdd_op = naive_sdd()

        self.add_operations(set_detector_op, autocor_op, naive_sdd_op)
        self.auto_connect_all()


class FourierCalibrationWorkflow(Workflow):
    def __init__(self):
        super(FourierCalibrationWorkflow, self).__init__('Fourier Calibration')

        autocor = fourier_autocorrelation()

        sdd = naive_sdd()

        self.add_operations(autocor, sdd)
        self.auto_connect_all()


class SimulateWorkflow(Workflow):
    def __init__(self):
        super(SimulateWorkflow, self).__init__('Calibrant Simulation')

        simulate = simulate_calibrant()

        self.add_operations(simulate)



from xicam.plugins.datahandlerplugin import DataHandlerPlugin, start_doc, descriptor_doc, event_doc, stop_doc, \
    embedded_local_event_doc

import os
import fabio
import uuid
import re
import functools
from pathlib import Path


class EDFPlugin(DataHandlerPlugin):
    name = 'EDFPlugin'

    DEFAULT_EXTENTIONS = ['.edf']

    descriptor_keys = ['ByteOrder', 'HeaderID', 'VersionNumber', 'Dim_1', 'Dim_2', 'count_time', 'object_keys']

    def __init__(self, path):
        super(EDFPlugin, self).__init__()
        self.path = path
        self.fimg = fabio.open(path)

    def __call__(self, *args, **kwargs):
        return self.fimg.data

    @staticmethod
    @functools.lru_cache(maxsize=10, typed=False)
    def parseTXTFile(path):
        p = Path(path)
        if not p.suffix == '.txt':
            path = str(p.with_suffix('.txt'))

        if not os.path.isfile(path):
            return dict()

        with open(path, 'r') as f:
            lines = f.readlines()

        paras = dict()

        # The 7.3.3 txt format is messy, with keyless values, and extra whitespaces

        keylesslines = 0
        for line in lines:
            cells = [_f for _f in re.split('[=:]+', line) if _f]

            key = cells[0].strip()

            if cells.__len__() == 2:
                cells[1] = cells[1].split('/')[0]
                paras[key] = key_cast(key, cells[1].strip())
            elif cells.__len__() == 1:
                keylesslines += 1
                paras['Keyless value #' + str(keylesslines)] = key

        return paras

    @staticmethod
    @functools.lru_cache(maxsize=10, typed=False)
    def parseDataFile(path):
        md = fabio.open(path).header
        md.update({'object_keys': {'pilatus2M': ['primary']}})
        return md


def key_cast(key, value):
    return conversions[key_type_map.get(key, 'str')](value)


_ALS_KEY_MAP = {
    'ABS(Vertical Beam Position)': 'event',
    'AI Channel 6': 'event',
    'AI Channel 7': 'event',
    'AIs': 'event',
    'AO Waveform': 'event',
    'Alpha_scan_I0_intensities': 'event',
    'Alpha_scan_I1_intensities': 'event',
    'Alpha_scan_diode_intensities': 'event',
    'Alpha_scan_positions': 'event',
    'Beam Current Over Threshold': 'event',
    'Beam Current': 'event',
    'Beamline Pass Beam AI': 'event',
    'Beamline Pass Beam': 'event',
    'Beamline Shutter AI': 'event',
    'Beamline Shutter Closed': 'event',
    'Beamline Shutter Open': 'event',
    'Beamstop X': 'event',
    'Beamstop Y': 'event',
    'Bruker pulses': 'event',
    'ByteOrder': ['start', 'event'],
    'DIOs': 'event',
    'DataType': ['start', 'event'],
    'Date': ['start', 'event'],
    'Detector Horizontal': 'event',
    'Detector Left Motor': 'event',
    'Detector Right Motor': 'event',
    'Detector Vertical': 'event',
    'Dim_1': ['descriptor', 'event'],
    'Dim_2': ['descriptor', 'event'],
    'EZ fast tension stage': 'event',
    'Exit Slit bottom': 'event',
    'Exit Slit left': 'event',
    'Exit Slit right': 'event',
    'Exit Slit top': 'event',
    'Feedback Interlock': 'event',
    'Flight Tube Horizontal': 'event',
    'Flight Tube Vertical': 'event',
    'GIWAXS beamstop X': 'event',
    'GIWAXS beamstop Y thorlabs': 'event',
    'GIWAXS beamstop Y': 'event',
    'Gate Shutter': 'event',
    'Gate': 'event',
    'GiSAXS Beamstop Counter': 'event',
    'GiSAXS Beamstop': 'event',
    'Hacked Ager Stage': 'event',
    'HeaderID': ['start', 'event'],
    'I1 AI': 'event',
    'I1': 'event',
    'Image': ['event', 'event'],
    'Izero AI': 'event',
    'Izero': 'event',
    'Keyless value #1': 'event',
    'Keyless value #2': 'event',
    'Keyless value #3': 'event',
    'Kramer strain data': 'event',
    'M1 Alignment Tune': 'event',
    'M1 Bend': 'event',
    'M1 Pitch': 'event',
    'M201 Feedback': 'event',
    'Mono Angle': 'event',
    'Motorized Lab Jack': 'event',
    'Motorized Lab Jack1': 'event',
    'Motors': 'event',
    'PCO Invert': 'event',
    'PHI Alignment Beamstop': 'event',
    'Pilatus 100K exp out': 'event',
    'Pilatus 1M Trigger Pulse': 'event',
    'Pilatus 300KW trigger pulse': 'event',
    'Printing motor': 'event',
    'SAXS Protector': 'event',
    'Sample Alpha Stage': 'event',
    'Sample Phi Stage': 'event',
    'Sample Rotation Stage ESP': 'event',
    'Sample Rotation Stage Miller': 'event',
    'Sample Rotation Stage': 'event',
    'Sample Thickness Stage': 'event',
    'Sample X Stage Fine': 'event',
    'Sample X Stage': 'event',
    'Sample Y Stage Arthur': 'event',
    'Sample Y Stage': 'event',
    'Sample Y Stage_old': 'event',
    'Size': ['descriptor', 'event'],
    'Slit 1 in Position': 'event',
    'Slit 2 in Position': 'event',
    'Slit Bottom Good': 'event',
    'Slit Top Good': 'event',
    'Slit1 bottom': 'event',
    'Slit1 left': 'event',
    'Slit1 right': 'event',
    'Slit1 top': 'event',
    'Sum of Slit Current': 'event',
    'Temp Beamline Shutter Open': 'event',
    'VersionNumber': ['start', 'event'],
    'Vertical Beam Position': 'event',
    'Xtal2 Pico 1 Feedback': 'event',
    'Xtal2 Pico 1': 'event',
    'Xtal2 Pico 2 Feedback': 'event',
    'Xtal2 Pico 2': 'event',
    'Xtal2 Pico 3 Feedback': 'event',
    'Xtal2 Pico 3': 'event',
    'count_time': ['descriptor', 'event'],
    'run': ['event', 'event'],
    'slit1 bottom current': 'event',
    'slit1 top current': 'event',
    'title': ['event', 'event'],
}

key_type_map = {'HeaderID': 'str',
                'Image': 'int',
                'VersionNumber': 'str',
                'ByteOrder': 'str',
                'DataType': 'str',
                'Dim_1': 'int',
                'Dim_2': 'int',
                'Size': 'int',
                'Date': 'date',
                'count_time': 'float',
                'title': 'str',
                'run': 'int',
                'Keyless value #1': 'float',
                'Keyless value #2': 'float',
                'Keyless value #3': 'float',
                'Motors': 'int',
                'Sample X Stage': 'float',
                'Sample Y Stage': 'float',
                'Sample Thickness Stage': 'float',
                'Sample X Stage Fine': 'float',
                'Sample Alpha Stage': 'float',
                'Sample Phi Stage': 'float',
                'M201 Feedback': 'float',
                'M1 Pitch': 'float',
                'Sample Rotation Stage': 'float',
                'M1 Bend': 'float',
                'Detector Horizontal': 'float',
                'Detector Vertical': 'float',
                'Slit1 top': 'float',
                'Slit1 bottom': 'float',
                'Slit1 right': 'float',
                'Slit1 left': 'float',
                'Exit Slit top': 'float',
                'Exit Slit bottom': 'float',
                'Exit Slit left': 'float',
                'Exit Slit right': 'float',
                'GIWAXS beamstop X': 'float',
                'GIWAXS beamstop Y': 'float',
                'Beamstop X': 'float',
                'Beamstop Y': 'float',
                'Detector Right Motor': 'float',
                'Detector Left Motor': 'float',
                'Motorized Lab Jack': 'float',
                'M1 Alignment Tune': 'float',
                'EZ fast tension stage': 'float',
                'Motorized Lab Jack1': 'float',
                'Sample Rotation Stage ESP': 'float',
                'Printing motor': 'float',
                'GIWAXS beamstop Y thorlabs': 'float',
                'Sample Y Stage Arthur': 'float',
                'Flight Tube Horizontal': 'float',
                'Flight Tube Vertical': 'float',
                'Hacked Ager Stage': 'float',
                'Sample Rotation Stage Miller': 'float',
                'Mono Angle': 'float',
                'Xtal2 Pico 1 Feedback': 'float',
                'Xtal2 Pico 2 Feedback': 'float',
                'Xtal2 Pico 3 Feedback': 'float',
                'Xtal2 Pico 1': 'float',
                'Xtal2 Pico 2': 'float',
                'Xtal2 Pico 3': 'float',
                'Sample Y Stage_old': 'float',
                'AO Waveform': 'float',
                'DIOs': 'int',
                'SAXS Protector': 'float',
                'Beamline Shutter Closed': 'float',
                'Beam Current Over Threshold': 'float',
                'Slit 1 in Position': 'float',
                'Slit 2 in Position': 'float',
                'Temp Beamline Shutter Open': 'float',
                'Beamline Shutter Open': 'float',
                'Feedback Interlock': 'float',
                'Beamline Pass Beam': 'float',
                'Gate Shutter': 'float',
                'Bruker pulses': 'float',
                'Slit Top Good': 'float',
                'Slit Bottom Good': 'float',
                'AIs': 'int',
                'Beam Current': 'float',
                'Beamline Shutter AI': 'float',
                'Beamline Pass Beam AI': 'float',
                'slit1 bottom current': 'float',
                'slit1 top current': 'float',
                'GiSAXS Beamstop': 'float',
                'Izero AI': 'float',
                'I1 AI': 'float',
                'PHI Alignment Beamstop': 'float',
                'AI Channel 6': 'float',
                'AI Channel 7': 'float',
                'Vertical Beam Position': 'float',
                'Pilatus 1M Trigger Pulse': 'float',
                'Pilatus 300KW trigger pulse': 'float',
                'PCO Invert': 'float',
                'Gate': 'float',
                'Izero': 'float',
                'I1': 'float',
                'GiSAXS Beamstop Counter': 'float',
                'Sum of Slit Current': 'float',
                'Pilatus 100K exp out': 'float',
                'Kramer strain data': 'float',
                'ABS(Vertical Beam Position)': 'float',
                'Alpha_scan_positions': 'tabdelimitedfloat',
                'Alpha_scan_I0_intensities': 'tabdelimitedfloat',
                'Alpha_scan_I1_intensities': 'tabdelimitedfloat',
                'Alpha_scan_diode_intensities': 'tabdelimitedfloat'
                }

conversions = {'int': lambda x: int(x.strip()),
               'float': lambda x: float(x.strip()),
               'str': lambda x: x.strip(),
               'date': lambda x: x.strip(),
               'tabdelimitedfloat': lambda x: list(map(float, x.split('\t'))) if x else []}


def _data_keys_from_value(v, src_name, object_name):
    kind_map = {'i': 'integer',
                'f': 'number',
                'U': 'string',
                'S': 'string'}
    return {'dtype': kind_map[np.array([v]).dtype.kind],
            'shape': [],
            'source': src_name,
            'object_name': object_name}


def _gen_descriptor_from_dict(ev_data, src_name):
    data_keys = {}
    confiuration = {}
    obj_keys = {}

    for k, v in ev_data.items():
        data_keys[k] = _data_keys_from_value(v, src_name, k)
        obj_keys[k] = [k]
        confiuration[k] = {'data': {},
                           'data_keys': {},
                           'timestamps': {}}

    return {'data_keys': data_keys,
            'time': time.time(),
            'uid': str(uuid.uuid4()),
            'configuration': confiuration,
            'object_keys': obj_keys}



from qtpy.QtCore import *
from qtpy.QtWidgets import *
from qtpy.QtGui import *
import pyqtgraph.parametertree.parameterTypes as pTypes
from pyqtgraph.parametertree import Parameter, ParameterTree, ParameterItem, registerParameterType
from pyFAI import calibrant
from collections import OrderedDict
from .workflows import FourierCalibrationWorkflow


class CalibrationPanel(ParameterTree):
    algorithms = OrderedDict(
        [('Fourier Autocorrelation', None),
         ('2D Ricker Wavelet', None),
         ('DPDAK Refinement', None)])
    sigCalibrate = Signal(object, str)
    sigDoCalibrateWorkflow = Signal(object)

    def __init__(self, headermodel: QStandardItemModel, selectionmodel: QItemSelectionModel):
        super(CalibrationPanel, self).__init__()

        self.selectionmodel = selectionmodel
        self.headermodel = headermodel

        self.header().close()

        calibrants = dict(zip(calibrant.ALL_CALIBRANTS.keys(), calibrant.ALL_CALIBRANTS.values()))
        self.device = pTypes.ListParameter(name='Device', type='list', values=[], value='')
        self.calibrant = pTypes.ListParameter(name='Calibrant Material', values=calibrants, value=calibrants['AgBh'])
        self.autoCalibrateMethod = pTypes.ListParameter(name='Algorithm', values=self.algorithms.keys())
        self.autoCalibrateAction = pTypes.ActionParameter(name='Auto Calibrate')
        self.autoCalibrateAction.sigActivated.connect(self.calibrate)
        self.calibratetext = pTypes.TextParameter(name='Instructions', value='', readonly=True, visible=False)

        self.parameter = pTypes.GroupParameter(name='Calibration', children=[self.device,
                                                                             self.calibrant,
                                                                             self.autoCalibrateMethod,
                                                                             self.autoCalibrateAction,
                                                                             self.autoCalibrateAction,
                                                                             self.calibratetext
                                                                             ])

        self.setParameters(self.parameter, showTop=False)

        self.workflow = FourierCalibrationWorkflow()

    def calibrate(self):
        self.sigDoCalibrateWorkflow.emit(self.workflow)

    def dataChanged(self, start, end, _):
        if not self.headermodel.itemFromIndex(self.selectionmodel.currentIndex()): return
        devices = self.headermodel.itemFromIndex(self.selectionmodel.currentIndex()).header.devices()
        self.parameter.param('Device').setLimits(list(set(devices)))
        self.parameter.param('Device').setValue(list(devices)[0])


from xicam.plugins import QWidgetPlugin
from pyqtgraph import ImageView, PlotItem
from xicam.core.data import NonDBHeader
from qtpy.QtWidgets import *
from qtpy.QtCore import *
from qtpy.QtGui import *
import numpy as np
from xicam.core import msg
from xicam.gui.widgets.tabview import TabView
from .SAXSViewerPlugin import SAXSViewerPluginBase


class SAXSMultiViewerPlugin(QSplitter, QWidgetPlugin):
    def __init__(self, headermodel, selectionmodel):
        super(SAXSMultiViewerPlugin, self).__init__()

        self.leftTabView = TabView()
        self.leftTabView.setWidgetClass(SAXSViewerPluginBase)
        self.leftTabView.setHeaderModel(headermodel)
        self.leftTabView.setSelectionModel(selectionmodel)
        self.rightTabView = TabView()
        self.rightTabView.setWidgetClass(SAXSViewerPluginBase)
        self.rightTabView.setHeaderModel(headermodel)

        self.addWidget(self.leftTabView)
        self.addWidget(self.rightTabView)

    def __getattr__(self, attr):  ## implicitly wrap methods from leftViewer
        if hasattr(self.plotwidget, attr):
            m = getattr(self.leftViewer, attr)
            if hasattr(m, '__call__'):
                return m
        raise NameError(attr)



from qtpy.QtWidgets import *
from qtpy.QtGui import *
from qtpy.QtCore import *
from xicam.plugins.widgetplugin import QWidgetPlugin
from pyqtgraph import PlotWidget, PlotDataItem, intColor, mkPen
from pyqtgraph.graphicsItems.LegendItem import ItemSample
import numpy as np
from xicam.gui.static import path
from xicam.core.execution.workflow import Workflow
from typing import Tuple


class SAXSSpectra(QTabWidget, QWidgetPlugin):
    name = 'SAXSSpectra'

    def __init__(self, workflow: Workflow, toolbar: QToolBar):
        super(SAXSSpectra, self).__init__()

        self._cache = {}  # cache is dict to allow future use of other keys as 'time' index
        self.workflow = workflow

        self.toolbar = toolbar
        self.toolbar.sigPlotCache.connect(self.replot_all)

        # self.legend = self.plotwidget.addLegend()

        # hbox = QHBoxLayout()
        # # hbox.addWidget(self.toolbar)
        # hbox.addWidget(self.plotwidget)
        # self.setLayout(hbox)
        # self.setContentsMargins(0, 0, 0, 0)
        # hbox.setSpacing(0)

    def setResult(self, result: Tuple[dict]):
        self.clear_all()
        self._cache[0] = tuple({k: v.value for k, v in r.items()} for r in result)
        self.replot_all()

    def appendResult(self, result):
        result_cache = tuple({k: v.value for k, v in r.items()} for r in result)
        self._cache[len(self._cache)] = result_cache
        self.plot_mode(result_cache)

    def plot_mode(self, resultset):
        self.clear()

        for result in resultset:
            name = next(iter(result.keys()))
            plotwidget = PlotWidget(
                labels={'bottom': 'q (\u212B\u207B\u00B9)', 'left': 'I (a.u.)', 'top': 'd (nm)'})

            def tickStrings(values, scale, spacing):
                return ['{:.3f}'.format(.2 * np.pi / i) if i != 0 else '\u221E' for i in values]

            plotwidget.plotItem.axes['top']['item'].tickStrings = tickStrings

            plotwidget.plot(*list(output.value for output in result.values()), name=name)

            self.addTab(plotwidget, name)

        #
        # checkedindices = self.toolbar.reductionModesModel.checkedIndices()
        # for name, xoutput, youtput in [
        #     (checkedindex.internalPointer().name, checkedindex.internalPointer().x, checkedindex.internalPointer().y)
        #     for checkedindex
        #     in checkedindices]:
        #     for result in resultset:
        #         if xoutput.name in result and youtput.name in result:
        #             self.plot(result[xoutput.name], result[youtput.name], name=name)

        # self._auto_pen()

    def plot(self, *args, **kwargs):
        self.plotwidget.plotItem.plot(*args, **kwargs)

    def _auto_pen(self):
        count = len(self.plotwidget.scene().items())
        self.clear_legend()
        for i, item in enumerate(self.plotwidget.scene().items()):
            if isinstance(item, PlotDataItem):
                item.setPen(mkPen(intColor(i, values=count), width=2))
                self.legend.addItem(item, item.name())

    def replot_all(self, checked=True):
        if not checked: return
        self.plotwidget.clear()
        for result in self._cache.values():
            self.plot_mode(result)

    def clear_all(self):
        self.plotwidget.clear()
        self.legend.items = []
        self._cache = {}

    def clear_legend(self):
        self.legend.scene().removeItem(self.legend)
        self.legend = self.plotwidget.addLegend()

    @property
    def inputs(self):
        return super(SAXSSpectra, self).inputs()

    def sizeHint(self):
        return QSize(150, 200)


class SAXSSpectraToolbar(QWidget):
    sigPlotCache = Signal()
    sigDoWorkflow = Signal()

    def __init__(self):
        super(SAXSSpectraToolbar, self).__init__()

        layout = QVBoxLayout()
        modetoolbar = QToolBar()
        optionstoolbar = QToolBar()
        layout.addWidget(modetoolbar)
        layout.addWidget(optionstoolbar)
        self.setLayout(layout)
        layout.setStretch(0, 2)
        layout.setStretch(1, 1)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        self.modeActionGroup = QActionGroup(self)
        qbtn = self.mkGroupToggle('icons/q.png', text='q (Azimuthal) Integration', receiver=self.sigPlotCache.emit)
        qbtn.setChecked(True)
        modetoolbar.addAction(qbtn)
        modetoolbar.addAction(
            self.mkGroupToggle('icons/chi.png', text='χ (chi/Radial) Integration', receiver=self.sigPlotCache.emit))
        modetoolbar.addAction(
            self.mkGroupToggle('icons/x.png', text='X (Horizontal) Integration', receiver=self.sigPlotCache.emit))
        modetoolbar.addAction(
            self.mkGroupToggle('icons/z.png', text='Z (Vertical) Integration', receiver=self.sigPlotCache.emit))
        modetoolbar.addAction(self.mkGroupToggle('icons/G.png', text='Guinier Plot', receiver=self.sigPlotCache.emit))
        modetoolbar.addAction(self.mkGroupToggle('icons/P.png', text='Porod Plot', receiver=self.sigPlotCache.emit))
        modetoolbar.addAction(self.mkGroupToggle('icons/Iq2.png', text='I×q\u00B2', receiver=self.sigPlotCache.emit))
        modetoolbar.addAction(self.mkGroupToggle('icons/Iq3.png', text='I×q\u00B3', receiver=self.sigPlotCache.emit))
        modetoolbar.addAction(self.mkGroupToggle('icons/Iq4.png', text='I×q\u0074', receiver=self.sigPlotCache.emit))
        modetoolbar.addAction(self.mkGroupToggle('icons/gofr.png', text='Electron Density Correlation Function',
                                                 receiver=self.sigPlotCache.emit))
        modetoolbar.addAction(
            self.mkGroupToggle('icons/gofrvec.png', text='Pair Distribution Function', receiver=self.sigPlotCache.emit))

        self.multiplot = QAction(self)
        self.multiplot.setIcon(QIcon(str(path('icons/multiplot.png'))))
        self.multiplot.setText('Plot Series')
        self.multiplot.setCheckable(True)
        self.multiplot.triggered.connect(self.sigDoWorkflow)
        optionstoolbar.addSeparator()
        optionstoolbar.addAction(self.multiplot)
        optionstoolbar.addAction(QIcon(str(path('icons/blackwhite.png'))), 'Toggle Theme')
        optionstoolbar.addAction(QIcon(str(path('icons/configure.png'))), 'Configure Plot')

        optionstoolbar.setOrientation(Qt.Vertical)
        modetoolbar.setOrientation(Qt.Vertical)

    def mkGroupToggle(self, iconpath: str = None, text=None, receiver=None):
        actn = QAction(self)
        if iconpath: actn.setIcon(QIcon(QPixmap(str(path(iconpath)))))
        if text: actn.setText(text)
        if receiver: actn.triggered.connect(receiver)
        actn.setCheckable(True)
        actn.setActionGroup(self.modeActionGroup)
        return actn



import time

import event_model

from xicam.core.execution import Workflow

from ..operations.average_intensity import average_intensity
from ..operations.fitting import fit_scattering_factor
from ..operations.fourierautocorrelator import fourier_correlation
from ..operations.onetime import one_time_correlation
from ..operations.twotime import two_time_correlation
from ..operations.correction import correct_fastccd_image
from ..operations.diffusion_coefficient import diffusion_coefficient


class ProcessingAlgorithms:
    """Convenience class to get the available algorithms that can be used for 1-time and 2-time correlations."""
    @staticmethod
    def algorithms():
        """Returns the available 1-time algorithms and 2-time algorithms."""
        return {
            TwoTimeAlgorithms.name: TwoTimeAlgorithms.algorithms(),
            OneTimeAlgorithms.name: OneTimeAlgorithms.algorithms()
        }

    @staticmethod
    def default():
        pass


class TwoTimeAlgorithms(ProcessingAlgorithms):
    name = '2-Time Algorithms'

    @staticmethod
    def algorithms():
        """Returns a dict where keys are the algorithm (workflow) names, values are the algorithms (workflows)."""
        return {TwoTime.name: TwoTime}

    @staticmethod
    def default():
        """Returns the default algorithm name to use."""
        return TwoTime.name


class OneTimeAlgorithms(ProcessingAlgorithms):
    name = '1-Time Algorithms'

    @staticmethod
    def algorithms():
        """Returns a dict where keys are the algorithm (workflow) names, values are the algorithms (workflows)."""
        return {OneTime.name: OneTime,
                FourierAutocorrelator.name: FourierAutocorrelator}

    @staticmethod
    def default():
        """Returns the default algorithm name to use."""
        return OneTime.name


class XPCSWorkflow(Workflow):
    def __init__(self):
        super(XPCSWorkflow, self).__init__()
        self.correct_image = correct_fastccd_image()
        self.add_operation(self.correct_image)


class TwoTime(XPCSWorkflow):
    name = '2-Time Correlation'

    def __init__(self):
        super(TwoTime, self).__init__()
        twotime = two_time_correlation()
        self.add_operation(twotime)
        self.add_link(self.correct_image, twotime, 'images', 'images')

    @staticmethod
    def document(**kwargs):
        results = kwargs.get('results')
        roi = kwargs.get('roi')
        workflow = kwargs.get('workflow_pickle')

        timestamp = time.time()

        run_bundle = event_model.compose_run()
        yield 'start', run_bundle.start_doc

        source = 'Xi-cam'

        peek_result = results[0]
        g2_shape = peek_result['g2'].value.shape[0]
        tau_shape = peek_result['tau'].value.shape[0]
        workflow = []
        workflow_shape = len(workflow)

        reduced_data_keys = {
            'norm-0-g2': {'source': source, 'dtype': 'number', 'shape': [g2_shape]},
            'tau': {'source': source, 'dtype': 'number', 'shape': [tau_shape]},
            'dqlist': {'source': source, 'dtype': 'string', 'shape': []},  # todo -- shape
            'workflow': {'source': source, 'dtype': 'string', 'shape': [workflow_shape]}
        }
        reduced_stream_name = 'reduced'
        reduced_stream_bundle = run_bundle.compose_descriptor(data_keys=reduced_data_keys,
                                                              name=reduced_stream_name)
        yield 'descriptor', reduced_stream_bundle.descriptor_doc

        # todo -- peek frame shape
        frame_data_keys = {'frame': {'source': source, 'dtype': 'number', 'shape': []}}
        frame_stream_name = 'primary'
        frame_stream_bundle = run_bundle.compose_descriptor(data_keys=frame_data_keys,
                                                            name=frame_stream_name)
        yield 'descriptor', frame_stream_bundle.descriptor_doc

        # todo -- store only paths? store the image data itself (memory...)
        # frames = header.startdoc['paths']
        frames = []
        for frame in frames:
            yield 'event', frame_stream_bundle.compose_event(
                data={frame},
                timestamps={timestamp}
            )

        for result in results:
            yield 'event', reduced_stream_bundle.compose_event(
                data={'norm-0-g2': result['g2'].value,
                      'tau': result['tau'].value,
                      'dqlist': roi,
                      'workflow': workflow},
                timestamps={'norm-0-g2': timestamp,
                            'tau': timestamp,
                            'dqlist': timestamp,
                            'workflow': workflow}
            )

        yield 'stop', run_bundle.compose_stop()


class OneTime(XPCSWorkflow):
    name = '1-Time Correlation'

    def __init__(self):
        super(OneTime, self).__init__()
        onetime = one_time_correlation()
        fitting = fit_scattering_factor()
        average_i = average_intensity()
        diffusion = diffusion_coefficient()

        self.add_operation(onetime)
        self.add_operation(average_i)
        self.add_operation(fitting)
        # self.add_operation(diffusion)

        # Manually set up connections (redundant if using this with LinearWorkflowEditor)
        self.add_link(self.correct_image, onetime, "images", "images")
        self.add_link(onetime, average_i, "images", "images")
        self.add_link(onetime, fitting, "g2", "g2")
        self.add_link(onetime, fitting, "tau", "tau")

    @staticmethod
    def document(**kwargs):
        results = kwargs.get('results')
        roi = kwargs.get('roi')
        workflow = kwargs.get('workflow_pickle')

        timestamp = time.time()

        run_bundle = event_model.compose_run()
        yield 'start', run_bundle.start_doc

        source = 'Xi-cam'

        peek_result = results[0]
        g2_shape = peek_result['g2'].value.shape[0]
        import numpy as np
        g2_err = np.zeros(g2_shape)
        g2_err_shape = g2_shape
        tau_shape = peek_result['tau'].value.shape[0]
        workflow = []
        workflow_shape = len(workflow)

        reduced_data_keys = {
            'norm-0-g2': {'source': source, 'dtype': 'number', 'shape': [g2_shape]},
            'norm-0-stderr': {'source': source, 'dtype': 'number', 'shape': [g2_err_shape]},
            'tau': {'source': source, 'dtype': 'number', 'shape': [tau_shape]},
            'g2avgFIT1': {'source': source, 'dtype': 'number', 'shape': [tau_shape]},
            'dqlist': {'source': source, 'dtype': 'string', 'shape': []},  # todo -- shape
            'workflow': {'source': source, 'dtype': 'string', 'shape': [workflow_shape]}
        }
        reduced_stream_name = 'reduced'
        reduced_stream_bundle = run_bundle.compose_descriptor(data_keys=reduced_data_keys,
                                                              name=reduced_stream_name)
        yield 'descriptor', reduced_stream_bundle.descriptor_doc

        # todo -- peek frame shape
        frame_data_keys = {'frame': {'source': source, 'dtype': 'number', 'shape': []}}
        frame_stream_name = 'primary'
        frame_stream_bundle = run_bundle.compose_descriptor(data_keys=frame_data_keys,
                                                            name=frame_stream_name)
        yield 'descriptor', frame_stream_bundle.descriptor_doc

        # todo -- store only paths? store the image data itself (memory...)
        # frames = header.startdoc['paths']
        frames = []
        for frame in frames:
            yield 'event', frame_stream_bundle.compose_event(
                data={frame},
                timestamps={timestamp}
            )

        for result in results:
            yield 'event', reduced_stream_bundle.compose_event(
                data={'norm-0-g2': result['g2'].value,
                      'norm-0-stderr': g2_err,
                      'tau': result['tau'].value,
                      'g2avgFIT1': result['fit_curve'].value,
                      'dqlist': roi,
                      'workflow': workflow},
                timestamps={'norm-0-g2': timestamp,
                            'norm-0-stderr': timestamp,
                            'tau': timestamp,
                            'g2avgFIT1': timestamp,
                            'dqlist': timestamp,
                            'workflow': workflow}
            )

        yield 'stop', run_bundle.compose_stop()


class FourierAutocorrelator(XPCSWorkflow):
    name = 'Fourier Correlation'

    def __init__(self):
        super(FourierAutocorrelator, self).__init__()
        fourier = FourierCorrelation()
        self.addProcess(fourier)



from xicam.plugins import QWidgetPlugin
from pyqtgraph import ImageView, PlotItem
from xicam.core.data import NonDBHeader
from qtpy.QtWidgets import *
from qtpy.QtCore import *
from qtpy.QtGui import *
import numpy as np
from xicam.core import msg
from xicam.plugins import manager as pluginmanager
from xicam.gui.widgets.dynimageview import DynImageView
from xicam.gui.widgets.imageviewmixins import Crosshair, QCoordinates, CenterMarker, BetterButtons, EwaldCorrected, \
    LogScaleIntensity, DisplayMode, CatalogView, XArrayView, ImageViewHistogramOverflowFix
import pyqtgraph as pg


# TODO LogScaleIntensity
class SAXSViewerPluginBase(CenterMarker, BetterButtons, Crosshair, QCoordinates, DynImageView, CatalogView, ImageViewHistogramOverflowFix):

    def __init__(self, *args, **kwargs):

        super(SAXSViewerPluginBase, self).__init__(*args, **kwargs)
        self.axesItem.invertY(False)

        # Setup coordinates label
        # self.coordinatesLbl = QLabel('--COORDINATES WILL GO HERE--')
        # self.ui.gridLayout.addWidget(self.coordinatesLbl, 3, 0, 1, 1, alignment=Qt.AlignHCenter)

        # Setup mask layer
        # TODO -- put in mixin
        self.maskimage = pg.ImageItem(opacity=.25, axisOrder='row-major')
        self.view.addItem(self.maskimage)

        # Setup calibration layer
        # TODO -- put in mixin
        self.calibrantimage = pg.ImageItem(opacity=.25)
        self.view.addItem(self.calibrantimage)

        # Empty ROI for later use
        # TODO -- refactor poly masking
        self.maskROI = pg.PolyLineROI([], closed=True, movable=False, pen=pg.mkPen(color='r', width=2))
        self.maskROI.handlePen = pg.mkPen(color='r', width=2)
        self.maskROI.handleSize = 10
        self.view.addItem(self.maskROI)

    def setMaskImage(self, mask):
        if mask is not None:
            self.maskimage.setImage(mask, lut=np.array([[0, 0, 0, 0], [255, 0, 0, 255]]))
            # self.maskimage.setTransform(QTransform(1, 0, 0, -1, 0, mask.shape[-2]))
        else:
            self.maskimage.clear()

    def setCalibrantImage(self, data):
        if data is not None:
            self.calibrantimage.setImage(data, lut=calibrantlut)
            # self.calibrantimage.setTransform(QTransform(0, 1, 1, 0, 0, 0))
        else:
            self.calibrantimage.clear()

    def redraw(self):
        if not self.parent().currentWidget() == self: return  # Don't redraw when not shown

        self.setHeader(self.header, self.field)

    def setResults(self, results):
        self.results = results
        self.redraw()


calibrantlut = np.array([[0, i, 0, i] for i in range(256)])


class SAXSCalibrationViewer(SAXSViewerPluginBase):
    pass


class SAXSMaskingViewer(SAXSViewerPluginBase):
    pass


class SAXSCompareViewer(SAXSViewerPluginBase):
    pass


class SAXSReductionViewer(EwaldCorrected, SAXSViewerPluginBase):
    def __init__(self, *args, toolbar: QToolBar = None, **kwargs):
        super(SAXSReductionViewer, self).__init__(*args, **kwargs)

        # Connect toolbar handlers
        self.toolbar = toolbar
        if self.toolbar:
            self.toolbar.modegroup.triggered.connect(self.setDisplayMode)
            # TODO -- re-connect the sigDeviceChanged outside of here
            # self.toolbar.sigDeviceChanged.connect(self.deviceChanged)
            # self.toolbar.sigDeviceChanged.connect(self.fieldChanged)


    def setDisplayMode(self, mode):
        if mode.text() == 'Wrap Ewald Sphere':
            mode = DisplayMode.remesh
        elif mode.text() == 'Raw':
            mode = DisplayMode.raw
        elif mode.text() == 'Cake (q/chi plot)':
            mode = DisplayMode.cake

        EwaldCorrected.setDisplayMode(self, mode)



import numpy as np
from astropy.modeling import fitting, models
from astropy.modeling import Fittable1DModel

from xicam.plugins.operationplugin import OperationPlugin

from typing import Tuple
from enum import Enum
from xicam.plugins import manager as pluginmanager


class QBackgroundFit(OperationPlugin):
    name = "Q Background Fit"

    def __init__(self):
        super(QBackgroundFit, self).__init__()
        self.peakranges = []

    def wireup_parameter(self, parameter):
        parameter.sigValueChanged.connect(self.value_changed)

    def value_changed(self, *args, **kwargs):
        print(f"QBackgroundFit.value_changed: {args} {kwargs}")
        self.find_peak_ranges()

    def find_peak_ranges(self):
        # FIXME: operation plugin has no reference to workflow..., how do we find peak ranges?
        print("find_peak_ranges")
        ...
        # must be a late import to avoid being picked up first by plugin manager
        # from xicam.SAXS.operations.astropyfit import AstropyQSpectraFit
        # thisindex = self._workflow.processes.index(self)
        # self.peakranges = [(process.domainmin.value, process.domainmax.value)
        #                    for process in self._workflow.processes[thisindex + 1:]
        #                    if isinstance(process, AstropyQSpectraFit)]

    def _func(self,
              q: np.ndarray,
              iq: np.ndarray,
              model: Enum,
              domain_min: float,
              domain_max: float,
              degree: int = 4):
        model = models.Polynomial1D(degree=degree)

        norange = domain_min == domain_max
        if domain_min is None and q is not None or norange:  # truncate the q and I arrays with limits
            domain_min = q.min()
        if domain_max is None and q is not None or norange:  # truncate the q and I arrays with limits
            domain_max = q.max()

        filter = np.logical_and(domain_min <= q, q <= domain_max)
        for peakrange in self.peakranges:
            print('applying peak range:', peakrange)
            filter &= np.logical_or(peakrange[0] >= q, q >= peakrange[1])

        q = q[filter]
        iq = iq[filter]
        background_model = fitting.LinearLSQFitter()(model, q, iq)
        background_profile = background_model.value(q)
        raw_iq = iq.copy()
        iq = iq - background_profile
        return q, iq, background_model, background_profile, raw_iq


# class QBackgroundFit(ProcessingPlugin):
#     name = 'Q Background Fit'
#
#     q = InOut(description='Q bin center positions',
#               type=np.array)
#     Iq = InOut(description='Q spectra bin intensities', type=np.array)
#     # model = Input(description='Fittable model class in the style of Astropy', type=Enum)
#     domainmin = Input(description='Min bound on the domain of the input data', type=float)
#     domainmax = Input(description='Max bound on the domain of the input data', type=float)
#     degree = Input(name='Polynomial Degree', description='Polynomial degree number', type=int, min=1, default=4)
#     # fitter = Input(description='Fitting algorithm', default=fitting.LevMarLSQFitter(), type=Enum, limits={'Linear LSQ':fitting.LinearLSQFitter(), 'Levenberg-Marquardt LSQ':fitting.LevMarLSQFitter(), 'SLSQP LSQ':fitting.SLSQPLSQFitter(), 'Simplex LSQ':fitting.SimplexLSQFitter()})
#     domainfilter = Input(description='Domain limits where peaks will be fitted; auto-populated by ')
#
#     backgroundmodel = Output(description='A new model with the fitted parameters; behaves as parameterized function',
#                              type=Fittable1DModel)
#     backgroundprofile = Output(description='The fitted profile from the evaluation of the '
#                                            'resulting model over the input range.')
#     rawIq = Output(description='The spectra data before subtraction.')
#
#     hints = [PlotHint(q, Iq), PlotHint(q, backgroundprofile), PlotHint(q, rawIq)]
#
#     modelvars = {}
#
#     def __init__(self):
#         super(QBackgroundFit, self).__init__()
#         self.peakranges = []
#
#     @property
#     def parameter(self):
#         self._workflow.attach(self.find_peak_ranges)  # order may be bad...
#         return super(QBackgroundFit, self).parameter
#
#     def find_peak_ranges(self):
#         from xicam.SAXS.operations.astropyfit import \
#             AstropyQSpectraFit  # must be a late import to avoid being picked up first by plugin manager
#         thisindex = self._workflow.processes.index(self)
#         self.peakranges = [(process.domainmin.value, process.domainmax.value)
#                            for process in self._workflow.processes[thisindex + 1:]
#                            if isinstance(process, AstropyQSpectraFit)]
#
#     def detach(self):
#         self._workflow.detach(self.find_peak_ranges)
#
#     def evaluate(self):
#         model = models.Polynomial1D(degree=self.degree.value)
#
#         norange = self.domainmin.value == self.domainmax.value
#         if self.domainmin.value is None and self.q.value is not None or norange:  # truncate the q and I arrays with limits
#             self.domainmin.value = self.q.value.min()
#         if self.domainmax.value is None and self.q.value is not None or norange:  # truncate the q and I arrays with limits
#             self.domainmax.value = self.q.value.max()
#
#         filter = np.logical_and(self.domainmin.value <= self.q.value, self.q.value <= self.domainmax.value)
#         for peakrange in self.peakranges:
#             print('applying peak range:', peakrange)
#             filter &= np.logical_or(peakrange[0] >= self.q.value, self.q.value >= peakrange[1])
#
#         q = self.q.value[filter]
#         Iq = self.Iq.value[filter]
#         self.backgroundmodel.value = fitting.LinearLSQFitter()(model, q, Iq)
#         self.backgroundprofile.value = self.backgroundmodel.value(self.q.value)
#         self.rawIq.value = self.Iq.value.copy()
#         self.Iq.value = self.Iq.value - self.backgroundprofile.value



from typing import Tuple

from xicam.core.intents import PlotIntent
from xicam.plugins.operationplugin import operation, output_names, display_name, describe_input, describe_output, \
    categories, intent
import numpy as np
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator


@operation
@output_names("q_z", "I")
@display_name("Z Integration")
@describe_input("azimuthal_integrator", 'A PyFAI.AzimuthalIntegrator object')
@describe_input("data", '2d array representing intensity for each pixel')
@describe_input("mask", 'Array (same size as image) with 1 for masked pixels, and 0 for valid pixels')
@describe_input("dark", "Dark noise image")
@describe_input("flat", "Flat field image")
@describe_input("normalization_factor", 'Value of normalization monitor')
@describe_output("q", 'q_z bin center positions')
@describe_output("I", "Binned/pixel-split integrated intensity")
@categories(("Scattering", "Integration"))
@intent(PlotIntent, name="Z Integration", output_map={'x': 'q_z', 'y': 'I'}, labels={"bottom": "q_z", "left": "I"})
def z_integrate(azimuthal_integrator: AzimuthalIntegrator,
                data: np.ndarray,
                mask: np.ndarray = None,
                dark: np.ndarray = None,
                flat: np.ndarray = None,
                normalization_factor: float = 1) -> Tuple[np.ndarray, np.ndarray]:
    if data.ndim == 2:
        data = [data]
    if dark is None: dark = np.zeros_like(data[0])
    if flat is None: flat = np.ones_like(data[0])
    if mask is None: mask = np.zeros_like(data[0])
    I = []
    q_z = []
    for frame in data:
        frame = np.asarray(frame)
        I_i = np.sum((frame - dark) * np.average(flat - dark) / (
                flat - dark) / normalization_factor * np.logical_not(mask), axis=-1)[::-1]
        I.append(I_i)
        centerx = azimuthal_integrator.getFit2D()['centerX']
        centerz = azimuthal_integrator.getFit2D()['centerY']
        q_z_i = azimuthal_integrator.qFunction(np.arange(0, frame.shape[-2]),
                                               np.array([centerx] * frame.shape[-2])) / 10
        q_z_i[np.arange(0, frame.shape[-2]) < centerz] *= -1.
        q_z_i = q_z_i[::-1]
        q_z.append(q_z_i)

        # TODO: support dynamic q

    return np.asarray(q_z_i), np.asarray(I)



from typing import Union, Tuple
import numpy as np
from pyFAI import units
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
from xicam.plugins.operationplugin import operation, output_names, display_name, describe_input, describe_output, \
    categories


@operation
@output_names('cake', 'chi', 'q')
@display_name("Cake Integrate")
@describe_input("azimuthal_integrator", 'A PyFAI.AzimuthalIntegrator object')
@describe_input("data", '2d array representing intensity for each pixel')
@describe_input("npt_rad", 'Number of bins along q')
@describe_input("npt_azim", 'Number of bins along chi')
@describe_input("polz_factor", 'Polarization factor for correction')
@describe_input("unit", 'Output units for q')
@describe_input("radial_range", 'The lower and upper range of the radial unit. If not provided, range is simply '
                                '(data.min(), data.max()). Values outside the range are ignored.')
@describe_input("azimuth_range", 'The lower and upper range of the azimuthal angle in degree. If not provided, '
                                 'range is simply (data.min(), data.max()). Values outside the range are ignored.')
@describe_input("mask", 'Array (same size as image) with 1 for masked pixels, and 0 for valid pixels')
@describe_input("dark", 'Dark noise image')
@describe_input("flat", 'Flat field image')
@describe_input("method", 'Can be "numpy", "cython", "BBox" or "splitpixel", "lut", "csr", "nosplit_csr", '
                          '"full_csr", "lut_ocl" and "csr_ocl" if you want to go on GPU. To Specify the device: '
                          '"csr_ocl_1,2"')
@describe_input("normalization_factor", 'Value of a normalization monitor')
@describe_output("chi", 'Chi bin center positions')
@describe_output("cake", 'Binned/pixel-split integrated intensity')
@describe_output("q", 'Q bin center positions')
@categories(("Scattering", "Transformations"))
def cake_integration(azimuthal_integrator: AzimuthalIntegrator,
                     data: np.ndarray,
                     npt_rad: int = 1000,
                     npt_azim: int = 1000,
                     polz_factor: float = 0,
                     unit: Union[str, units.Unit] = "q_A^-1",
                     radial_range: Tuple[float] = None,
                     azimuth_range: Tuple[float] = None,
                     mask: np.ndarray = None,
                     dark: np.ndarray = None,
                     flat: np.ndarray = None,
                     method: str = 'splitbbox',
                     normalization_factor: float = 1) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    cake, q, chi = azimuthal_integrator.integrate2d(data=data,
                                                    npt_rad=npt_rad,
                                                    npt_azim=npt_azim,
                                                    radial_range=radial_range,
                                                    azimuth_range=azimuth_range,
                                                    mask=mask,
                                                    polarization_factor=polz_factor,
                                                    dark=dark,
                                                    flat=flat,
                                                    method=method,
                                                    unit=unit,
                                                    normalization_factor=normalization_factor)

    return cake, q, chi

# # TODO: check if the data actually needs to be flipped ud; keeping this for posterity
# def nonesafe_flipud(data: np.ndarray):
#     if data is None: return None
#     return np.flipud(data).copy()



from typing import Union, Tuple

from xicam.core.intents import PlotIntent
from xicam.plugins.operationplugin import operation, output_names, display_name, describe_input, describe_output, \
    categories, intent
import numpy as np
from pyFAI import units
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator


@operation
@output_names("chi", "I")
@display_name("Chi Integration")
@describe_input("azimuthal_integrator", 'A PyFAI.AzimuthalIntegrator object')
@describe_input("data", '2d array representing intensity for each pixel')
@describe_input("npt_azim", 'Number of bins along chi')
@describe_input("polz_factor", 'Polarization factor for correction')
@describe_input("unit", 'Output units for q')
@describe_input("radial_range", 'The lower and upper range of the radial unit. If not provided, range is simply '
                                '(data.min(), data.max()). Values outside the range are ignored.')
@describe_input("azimuth_range", 'The lower and upper range of the azimuthal angle in degree. If not provided, '
                                 'range is simply (data.min(), data.max()). Values outside the range are ignored.')
@describe_input("mask", 'Array (same size as image) with 1 for masked pixels, and 0 for valid pixels')
@describe_input("dark", 'Dark noise image')
@describe_input("flat", 'Flat field image')
@describe_input("method", 'Can be "numpy", "cython", "BBox" or "splitpixel", "lut", "csr", "nosplit_csr", '
                          '"full_csr", "lut_ocl" and "csr_ocl" if you want to go on GPU. To Specify the device: '
                          '"csr_ocl_1,2"')
@describe_input('normalization_factor', 'Value of a normalization monitor')
@describe_output("chi", 'Q bin center positions')
@describe_output("I", 'Binned/pixel-split integrated intensity')
@intent(PlotIntent, name="Chi Integrate", output_map={"x": "chi", "y": "I"}, labels={"bottom": "chi", "left": "I"})
@categories(("Scattering", "Integration"))
def chi_integrate(azimuthal_integrator: AzimuthalIntegrator,
                  data: np.ndarray,
                  npt_azim: int = 1000,
                  polz_factor: float = 0,
                  unit: Union[str, units.Unit] = "q_A^-1",
                  radial_range: Tuple[float, float] = None,
                  azimuth_range: Tuple[float, float] = None,
                  mask: np.ndarray = None,
                  dark: np.ndarray = None,
                  flat: np.ndarray = None,
                  method: str = 'splitbbox',
                  normalization_factor: float = 1, ) -> Tuple[np.ndarray, np.ndarray]:
    I = []
    chi = []
    if data.ndim == 2:
        data = [data]
    for frame in data:
        I_i, q_i, chi_i = azimuthal_integrator.integrate2d(data=np.asarray(frame),
                                                           npt_rad=1,
                                                           npt_azim=npt_azim,
                                                           radial_range=radial_range,
                                                           azimuth_range=azimuth_range,
                                                           mask=mask,
                                                           polarization_factor=polz_factor,
                                                           dark=dark,
                                                           flat=flat,
                                                           method=method,
                                                           unit=unit,
                                                           normalization_factor=normalization_factor)

        I_i = np.sum(I_i, axis=1)
        chi.append(chi_i)
        I.append(I_i)
    return np.asarray(chi_i), np.asarray(I)  # TODO: support dynamic q

# def nonesafe_flipud(data: np.ndarray):
#     if data is None: return None
#     return np.flipud(data).copy()



# from xicam.SAXS.patches.pyFAI import *
import dill
import pyFAI
# from distributed.protocol.serialize import serialize as dumps, deserialize as loads
from pickle import dumps, loads

def test_Detector_pickle():
    import cloudpickle
    import numpy as np
    from pyFAI.detectors import Pilatus2M
    det = Pilatus2M()

    print(det, type(det))

    # print(det.__reduce__())
    # print(det.__getnewargs_ex__())
    # print(det.__getstate__())

    assert dumps(det)
    assert loads(dumps(det))
    assert loads(dumps(det)).shape == (1679, 1475)


def test_AzimuthalIntegrator_pickle():
    import dill
    import numpy as np
    from pyFAI.azimuthalIntegrator import AzimuthalIntegrator

    det = pyFAI.detectors.detector_factory('pilatus2m')
    ai = AzimuthalIntegrator(detector=det)
    ai.set_wavelength(.1)
    spectra = ai.integrate1d(np.ones(det.shape), 1000)  # force lut generation
    dump = dumps(ai)
    newai = loads(dump)
    assert np.array_equal(newai.integrate1d(np.ones(det.shape), 1000), spectra)
    assert newai.detector.shape == (1679, 1475)


def test_Calibrant():
    from pyFAI import calibrant
    calibrant = calibrant.CalibrantFactory()('AgBh')
    assert dumps(calibrant)
    assert loads(dumps(calibrant))



from databroker import Broker
from qtpy.QtWidgets import QPushButton, QSizePolicy
import numpy as np

from xicam.gui.widgets.imageviewmixins import BetterLayout, ProcessingView

from xicam.SAXS.operations.correction import correct


class BackgroundCorrected(BetterLayout, ProcessingView):
    """Toggle background correction for an image or image series."""
    def __init__(self, *args, darks=None, **kwargs):
        self._darks = None
        self.set_darks(darks)
        self._bg_correction = False
        self._bg_correct_btn = QPushButton("BG Correction")
        self._bg_correct_btn.setEnabled(False)
        self._bg_correct_btn.setCheckable(True)
        self._bg_correct_btn.clicked.connect(self._toggle_bg_correction)

        super(BackgroundCorrected, self).__init__(*args, **kwargs)

        size_policy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        size_policy.setHorizontalStretch(0)
        size_policy.setVerticalStretch(1)
        size_policy.setHeightForWidth(self._bg_correct_btn.sizePolicy().hasHeightForWidth())
        self._bg_correct_btn.setSizePolicy(size_policy)
        self.ui.right_layout.addWidget(self._bg_correct_btn)

    def setImage(self, img, *args, darks=None, **kwargs):
        """Override to add additional `darks` kwarg for dark image."""
        if darks is not None and self._darks is not None:
            self._darks = darks
        self.set_darks(darks)  # Can't do this before... self.image won't be set until super call
        super(BackgroundCorrected, self).setImage(img, *args, **kwargs)

    def set_darks(self, darks):
        """Set dark image for this image view."""
        if darks is not None:
            self._darks = darks
            if self._darks.ndim == 3:
                self._darks = np.mean(self._darks, axis=0)
            self._bg_correct_btn.setEnabled(True)
            self._toggle_bg_correction(True)

    def _toggle_bg_correction(self, value):
        """Slot to handle when bg correction button is clicked."""
        self._bg_correction = value
        self._bg_correct_btn.setChecked(value)
        index = self.currentIndex
        self.setImage(self.image, autoRange=False)  # Reprocess the current frame
        if self.image is not None:
            try:
                self.setCurrentIndex(index)
            except TypeError as e:
                # Ignore "tuple indices must be integers or slices, not NoneType" for 2d pseudo3framearray images
                pass

    def process(self, image):
        """Either returns the raw image or the correct image, depending on the bg correction button state."""
        if self._bg_correction:
            flats = np.ones_like(image)
            if self._darks is not None:
                image = correct(np.expand_dims(image, 0), flats, self._darks)[0]
        return super(BackgroundCorrected, self).process(image)



import numpy as np
from xicam.SAXS.utils import get_label_array

from xicam.plugins.operationplugin import intent, operation, output_names, visible
from xicam.core.intents import PlotIntent


@operation
@output_names("times", "intensities", "images")
@intent(PlotIntent,
        "Average Intensity",
        output_map={"x": "times", "y": "intensities"},
        labels={"bottom": "t", "left": "I"},
        mixins=["ToggleSymbols"])
@visible('images', False)
@visible('rois', False)
def average_intensity(images: np.ndarray, labels: np.ndarray = None, rois: np.ndarray = None):
    # TODO: calculate labels from ROIs

    if labels is None and rois is not None:
        labels = get_label_array(images, rois=rois)

        # Trim the image based on labels, and resolve to memory
        si, se = np.where(np.flipud(labels))
        images = np.asarray(images[:, si.min():si.max() + 1, se.min():se.max() + 1])
        labels = np.asarray(np.flipud(labels)[si.min():si.max() + 1, se.min():se.max() + 1])

    averages = []
    if labels is not None:
        n_labels = int(labels.max())
        for label in range(1, n_labels + 1):
            masked_image = np.where(labels == label, images, np.zeros(images[0].shape))
            averages.append(np.average(masked_image, axis=(-2, -1)))
    else:
        averages = np.average(trimmed_images, axis=(-2, -1))

    return np.asarray(range(len(images))), np.asarray(averages).squeeze(), images
